#from cosymlib.molecule.geometry import Geometry
from cosymlib.molecule.electronic_structure import ElectronicStructure
from cosymlib.simulation import ExtendedHuckel
from cosymlib.tools import element_to_atomic_number
from warnings import warn
from copy import deepcopy


# Gets the parameters defined in the arguments of the function and sets them to Symmetry instance
def set_parameters(func):
    def wrapper(*args, **kwargs):
        args[0]._symmetry.set_parameters(kwargs)
        args[0]._symmetry.set_electronic_structure(args[0].electronic_structure)
        return func(*args, **kwargs)
    return wrapper


class Molecule:
    """
    Main molecule class

    :param geometry: The geometry
    :type geometry: Geometry
    :param electronic_structure: The electronic structure
    :type electronic_structure: ElectronicStructure
    :param name: Molecule name
    :type name: str

    """
    def __init__(self, geometry,
                 electronic_structure=None,
                 name=None):

        if not geometry:
            raise Exception('No geometry found in the input file, check out input file for possible errors')
        if name is None:
            self._name = geometry.name

        self._geometry = geometry
        self._electronic_structure = electronic_structure
        self._symmetry = geometry._symmetry
        self._shape = geometry._shape

    def get_copy(self):
        return deepcopy(self)

    def set_name(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def geometry(self):
        """
        Get the geometry

        :return: The geometry
        :rtype: Geometry
        """
        return self._geometry

    @property
    def electronic_structure(self):
        """
        Get the electronic structure

        :return: The electronic structure
        :rtype: ElectronicStructure
        """
        if self._electronic_structure is None:
            warn('Electronic structure auto generated from Extended-Huckel calculation')
            eh = ExtendedHuckel(self.geometry)
            self._electronic_structure = ElectronicStructure(basis=eh.get_basis(),
                                                             orbital_coefficients=[eh.get_mo_coefficients(), []],
                                                             alpha_energies=eh.get_mo_energies(),
                                                             beta_energies=[],
                                                             multiplicity=eh.get_multiplicity(),
                                                             alpha_occupancy=[1]*eh.get_alpha_electrons(),
                                                             beta_occupancy=[1]*eh.get_beta_electrons())

        return self._electronic_structure

    # TODO: 'symmetry' and 'shape' properties may be removed. All methods inside these
    # TODO: classes may be mirrored in geometry/molecule class? [idea]
    @property
    def symmetry(self):
        return self._symmetry

    @property
    def shape(self):
        return self._shape

    # New ones (to substitute get_mo_symmetry)
    @set_parameters
    def get_mo_irreducible_representations(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.mo_irreducible_representations(group)

    @set_parameters
    def get_wf_irreducible_representations(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.wf_irreducible_representations(group)

    @set_parameters
    def get_mo_overlaps(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.mo_overlaps(group)

    @set_parameters
    def get_wf_overlaps(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.wf_overlaps(group)

    @set_parameters
    def get_symmetry_matrix(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.symmetry_matrix(group)

    @set_parameters
    def get_wf_symmetry(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.wf_measure(group)

    @set_parameters
    def get_dens_symmetry(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.dens_measure(group)

    @set_parameters
    def get_symmetry_axes(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.axes(group)

    @set_parameters
    def get_ideal_group_table(self, group, axis=None, axis2=None, center=None):
        return self._symmetry.wf_ideal_group_table(group)

    # Mirror methods in geometry
    def get_connectivity(self):
        return self.geometry.get_connectivity()

    def get_positions(self):
        """
        Get the positions in Cartesian coordinates

        :return: the coordinates
        :rtype: list
        """
        return self.geometry.get_positions()

    def get_n_atoms(self):
        """
        Get the number of atoms

        :return: number of atoms
        :rtype: int
        """
        return len(self.get_positions())

    def get_symbols(self):
        """
        Get the atomic elements symbols

        :return: the symbols
        :rtype: list
        """
        return self.geometry.get_symbols()

    def get_pointgroup(self, tol=0.01):
        """
        Get the symmetry point group

        :param tol: The tolerance
        :type tol: float
        :return: The point group
        :rtype: str
        """
        return self.geometry.get_point_group(tol=tol)

    def get_charge(self):
        """
        Get the charge of the molecule
        """
        if self._electronic_structure is None:
            return None
        net_electrons = sum([element_to_atomic_number(symbol) for symbol in self.get_symbols()])
        return net_electrons - (sum(self.electronic_structure.alpha_occupancy) +
                                sum(self.electronic_structure.beta_occupancy))



__version__ = '0.9.5'

from cosymlib.molecule import Molecule
from cosymlib.molecule.geometry import Geometry
from cosymlib import file_io
from cosymlib import tools
from cosymlib.utils import swap_vectors
from cosymlib.utils import get_shape_path, plot_molecular_orbital_diagram, plot_symmetry_energy_evolution
from cosymlib.shape.tools import get_structure_references, get_sym_from_label
import matplotlib.pyplot as plt
from warnings import warn

import sys
import os
import numpy as np


def _get_symmetry_arguments(locals):
    kwargs = dict(locals)
    del kwargs['self']
    for element in ['self', 'output']:
        if element in kwargs:
            del kwargs[element]

    return kwargs


def _get_table_format(labels, molecules_names, data):

    txt = 'Structure'
    max_len_name = 12
    max_name = 0
    names = []
    for name in molecules_names:
        if len(name) > max_len_name:
            max_name = 12
            names.append(name[:(max_len_name - 1)])
        elif len(name) > max_name:
            max_name = len(name)
            names.append(name)

    n = max_len_name - 3
    for label in labels:
        n += len(label)
        txt += '{}'.format(label.rjust(n))
        n = 11 - len(label)
    txt += '\n\n'

    for idx, name in enumerate(names):
        txt += '{:{width}}'.format(name + ',', width=max_len_name)
        n = 11
        for idn, label in enumerate(labels):
            txt += '{:>{width}.{prec}f},'.format(data[idn][idx], width=n, prec=3)
            n = 10
        txt += '\n'
    txt += '\n'
    return txt


def _get_axis_info(molecule, group, axis, axis2, center):

    txt = ''
    axis_info = molecule.get_symmetry_axes(group, axis=axis, axis2=axis2, center=center)
    txt += '\nSymmetry axis orientation\n'
    txt += 'center: ' + '  '.join(['{:12.8f}'.format(s) for s in axis_info['center']])
    txt += '\n'
    txt += 'axis  : ' + '  '.join(['{:12.8f}'.format(s) for s in axis_info['axis']])
    txt += '\n'
    txt += 'axis2 : ' + '  '.join(['{:12.8f}'.format(s) for s in axis_info['axis2']])
    txt += '\n'
    return txt


class Cosymlib:
    """
    Main cosymlib class

    :param structures: List of :class:`~cosymlib.molecule.geometry.Geometry` or :class:`~cosymlib.molecule.Molecule`
    type objects
    :type structures: list
    :param ignore_atoms_labels: Ignore atomic element labels is symmetry calculations
    :type ignore_atoms_labels: bool
    :param ignore_connectivity: Ignore connectivity in symmetry calculations
    :type ignore_connectivity: bool
    :param connectivity: List of pairs if atom indices that are considered connected
    :type connectivity: list
    :param connectivity_thresh: Connectivity threshold (Ionic radius is used as reference)
    :type connectivity_thresh: bool

    """

    def __init__(self,
                 structures,
                 ignore_atoms_labels=False,
                 ignore_connectivity=False,
                 connectivity=None,
                 connectivity_thresh=None):

        self._molecules = []
        if isinstance(structures, list):
            for structure in structures:
                if isinstance(structure, Molecule):
                    self._molecules.append(structure)
                elif isinstance(structure, Geometry):
                    self._molecules.append(Molecule(structure))
                else:
                    raise AttributeError('Molecule object not found')
        else:
            if isinstance(structures, Molecule):
                self._molecules.append(structures)
            elif isinstance(structures, Geometry):
                self._molecules.append(Molecule(structures))
            else:
                raise AttributeError('Molecule object not found')

        for molecule in self._molecules:
            if ignore_atoms_labels:
                molecule.geometry.set_symbols('X' * molecule.geometry.get_n_atoms())
            if connectivity_thresh is not None:
                molecule.geometry.generate_connectivity(thresh=connectivity_thresh)
            if connectivity is not None:
                molecule.geometry.set_connectivity(connectivity)
            if ignore_connectivity:
                molecule.geometry.set_connectivity(None)

    def get_n_atoms(self):
        """
        Get the number of atoms if all structures contains the same number of atoms,
        else raise exception.

        :return: Number of atoms
        :rtype: int
        """
        n_atoms_unique_list = np.unique([mol.geometry.get_n_atoms() for mol in self._molecules])
        if len(n_atoms_unique_list) > 1:
            raise Exception('Not all structures have same number of atoms')

        return n_atoms_unique_list[0]

    def get_geometries(self):
        """
        Get the geometries

        :return: List of :class:`~cosymlib.molecule.geometry.Geometry` objects
        :rtype: list
        """
        return [mol.geometry for mol in self._molecules]

    def print_info(self):
        """
        Prints general information about the structures
        """
        print('\033[1m{:20}   {:^5}\033[0m'.format('name', 'atoms'))
        for molecule in self._molecules:
            print('{:20} : {:5}'.format(molecule.name, molecule.geometry.get_n_atoms()))
        print('Total structures: {}'.format(len(self._molecules)))

    @property
    def molecules(self):
        """
        Get the molecules

        :return: List of :class:`~cosymlib.molecule.Molecule` objects
        :rtype: list
        """
        return self._molecules

    def print_shape_measure(self, shape_reference, central_atom=0, fix_permutation=False, output=sys.stdout):
        """
        Prints the shape measure of all structures in format

        :param shape_reference: List of references and/or geometries
        :type shape_reference: list
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param fix_permutation: Do not permute atoms during shape calculations
        :type fix_permutation: bool
        :param output: Display hook
        :type output: hook
        """

        molecules_names = [molecule.name for molecule in self._molecules]

        measure_list = []
        references_names = []
        for reference in shape_reference:
            measure_list.append(self.get_shape_measure(reference,
                                                       'measure',
                                                       central_atom,
                                                       fix_permutation))

            if type(reference) is Geometry:
                references_names.append(reference.name)
            else:
                references_names.append(reference)

        txt_shape = _get_table_format(references_names, molecules_names, measure_list)
        output.write(txt_shape)

    def print_shape_structure(self, shape_reference, central_atom=0, fix_permutation=False, output=sys.stdout):
        """
        Prints the nearest shape structure in format

        :param shape_reference: List of references and/or geometries
        :type shape_reference: list
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param fix_permutation: Do not permute atoms during shape calculations
        :type fix_permutation: bool
        :param output: Display hook
        :type output: hook
        """

        shape_results_structures = []
        references = []
        sym_labels = []
        for reference in shape_reference:
            if isinstance(reference, str):
                references.append(reference)
                sym_labels.append(get_sym_from_label(reference))
            else:
                references.append(reference.name)
                sym_labels.append('')

            shape_results_structures.append(self.get_shape_measure(reference,
                                                                   'structure',
                                                                   central_atom,
                                                                   fix_permutation))
        geometries = []
        for idm, molecule in enumerate(self._molecules):
            geometries.append(molecule.geometry)
            for idl, reference in enumerate(references):
                shape_results_structures[idl][idm].set_name(molecule.name + ' ' + reference + ' ' +
                                                            sym_labels[idl])
                geometries.append(shape_results_structures[idl][idm])

        print("\nOriginal structures vs reference polyhedra in file {}\n".format(output.name))
        for geometry in geometries:
            output.write(file_io.get_file_xyz_txt(geometry))

    def print_geometric_measure_info(self, label, multi=1, central_atom=0, center=None, output=sys.stdout):
        """
        Prints geometric symmetry measure verbose

        :param label: Symmetry point group label
        :type label: str
        :param multi: Number of symmetry axis to find
        :type multi: int
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param center: Center of symmetry in Cartesian coordinates. If None center is optimized
        :type center: list
        :param output: Display hook
        :type: hook
        """
        kwargs = _get_symmetry_arguments(locals())
        sep_line = '..................................................\n'

        txt = 'Evaluating symmetry operation : {}\n'.format(label)

        for idx, molecule in enumerate(self._molecules):
            molecule.geometry._symmetry.set_parameters(kwargs)
            txt += '{}\n'.format(molecule.name)
            txt += '\n'
            txt += 'Centered Structure\n'
            txt += sep_line
            cm = tools.center_mass(molecule.geometry.get_symbols(), molecule.geometry.get_positions())
            for idn, array in enumerate(molecule.geometry.get_positions()):
                array = array - cm
                txt += '{:2} {:12.8f} {:12.8f} {:12.8f}\n'.format(molecule.geometry.get_symbols()[idn],
                                                                  array[0], array[1], array[2])
            txt += sep_line

            txt += 'Optimal permutation\n'
            for idn, permutation in enumerate(molecule.geometry._symmetry.optimum_permutation(label)):
                txt += '{:2} {:2}\n'.format(idn + 1, permutation)
            txt += '\n'

            txt += 'Inverted structure\n'
            for idn, axis in enumerate(molecule.geometry._symmetry.nearest_structure(label)):
                txt += '{:2} {:12.8f} {:12.8f} {:12.8f}\n'.format(molecule.geometry.get_symbols()[idn],
                                                                  axis[0], axis[1], axis[2])
            txt += '\n'

            txt += 'Reference axis\n'
            for array in molecule.geometry._symmetry.reference_axis(label):
                txt += '{:12.8f} {:12.8f} {:12.8f}\n'.format(array[0], array[1], array[2])
            txt += '\n'
            txt += 'Symmetry measure {:.5f}\n'.format(molecule.geometry.get_symmetry_measure(**kwargs))
            txt += sep_line

        output.write(txt)

    def print_geometric_symmetry_measure(self, label, central_atom=0, center=None, output=sys.stdout):
        """
        Prints geometric symmetry measure in format

        :param label: Symmetry point group label
        :type label: str
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param center: Center of symmetry in Cartesian coordinates. If None center is optimized
        :type center: list
        :param output: Display hook
        :type output: hook
        """
        kwargs = _get_symmetry_arguments(locals())

        txt = 'Evaluating symmetry operation : {}\n \n'.format(label)
        for idx, molecule in enumerate(self._molecules):
            csm = molecule.geometry.get_symmetry_measure(**kwargs)
            max_name = len(max(molecule.name, key=len))
            txt += '{} '.format(molecule.name)
            if max_name < 9:
                n = 18 - len(molecule.name)
            else:
                n = 9 + max_name - len(molecule.name)
            txt += '{:{width}.{prec}f}\n'.format(csm, width=n, prec=3)

        output.write(txt)

    def print_symmetry_nearest_structure(self, label, central_atom=0, center=None, output=sys.stdout):
        """
        Prints the nearest structure to ideal symmetric structure

        :param label: Symmetry point group label
        :type label: str
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param center: Center of symmetry in Cartesian coordinates. If None center is optimized
        :type center: int
        :param output: Display hook
        :type output: hook
        """
        kwargs = _get_symmetry_arguments(locals())

        for idm, molecule in enumerate(self._molecules):
            geometry = molecule.geometry.get_symmetry_nearest_structure(**kwargs)
            output.write(file_io.get_file_xyz_txt(geometry))

    def print_chirality_measure(self, labels, central_atom=0, center=None, output=sys.stdout):

        csm_list = []
        for label in labels:
            csm_list.append([geometry.get_symmetry_measure(label, central_atom=central_atom, center=center)
                             for geometry in self.get_geometries()])

        molecules_names = [molecule.name for molecule in self._molecules]
        txt = _get_table_format(labels, molecules_names, csm_list)

        output.write(txt)

    def _print_electronic_symmetry_measure(self, group, axis=None, axis2=None, center=None, output=sys.stdout):

        txt = ''
        first = True
        for molecule in self._molecules:
            wf_measure = molecule.get_wf_symmetry(group, axis=axis, axis2=axis2, center=center)

            if first:
                sep_line = '          ' + '---------' * len(wf_measure['labels']) + '\n'

                txt += '\nWaveFunction: CSM-like values\n'
                txt += sep_line
                txt += '           ' + '  '.join(['{:^7}'.format(s) for s in wf_measure['labels']])
                txt += '\n'
                txt += sep_line

            txt += '{:<9} '.format(molecule.name) + '  '.join(['{:7.3f}'.format(s) for s in wf_measure['csm']])
            txt += '\n'
            first = False
        axes_information = molecule.get_symmetry_axes(group, axis=axis, axis2=axis2, center=center)
        txt += '\nSymmetry parameters\n'
        txt += 'center: ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['center']])
        txt += '\n'
        txt += 'axis  : ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['axis']])
        txt += '\n'
        txt += 'axis2 : ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['axis2']])
        txt += '\n'

        output.write(txt)

    def print_electronic_density_measure(self, group, axis=None, axis2=None, center=None, output=sys.stdout):

        txt = ''
        txt2 = ''
        first = True
        for molecule in self._molecules:
            dens_measure = molecule.get_dens_symmetry(group, axis=axis, axis2=axis2, center=center)

            if first:
                sep_line = '          ' + '---------' * len(dens_measure['labels']) + '\n'

                txt += '\nDensity: CSM-like values\n'
                txt += sep_line
                txt += '           ' + '  '.join(['{:^7}'.format(s) for s in dens_measure['labels']])
                txt += '\n'
                txt += sep_line

                txt2 += '--------------\n'
                txt2 += 'Total CSM {}\n'.format(group)
                txt2 += '--------------\n'

            txt += '{:<9} '.format(molecule.name) + '  '.join(['{:7.3f}'.format(s) for s in dens_measure['csm_coef']])
            txt += '\n'
            first = False
            axes_information = molecule.get_symmetry_axes(group, axis=axis, axis2=axis2, center=center)
            txt2 += '{:<9} '.format(molecule.name) + '{:7.3f}\n'.format(dens_measure['csm'])
            txt2 += '\nSymmetry parameters\n'
            txt2 += 'center: ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['center']])
            txt2 += '\n'
            txt2 += 'axis  : ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['axis']])
            txt2 += '\n'
            txt2 += 'axis2 : ' + '  '.join(['{:12.8f}'.format(s) for s in axes_information['axis2']])
            txt2 += '\n'

        txt += '\n' + txt2

        output.write(txt)

    def print_wnfsym_sym_matrices(self, group, axis=None, axis2=None, center=None, output=sys.stdout):

        txt = ''
        for molecule in self._molecules:

            sym_mat = molecule.get_symmetry_matrix(group, axis=axis, axis2=axis2, center=center)

            geometry = molecule.geometry
            sym_lables = sym_mat['labels']
            sym_matrices = sym_mat['matrix']

            sep_line = '--------------------------------------------\n'
            txt += 'MEASURES OF THE SYMMETRY GROUP:   {}\n'.format(group)
            txt += 'Basis: {}\n'.format(list(molecule.electronic_structure.basis.keys())[0])
            txt += sep_line
            txt += ' Atomic Coordinates (Angstroms)\n'
            txt += sep_line
            for idn, array in enumerate(geometry.get_positions()):
                txt += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idn],
                                                                  array[0], array[1], array[2])
            txt += sep_line
            for i, group in enumerate(sym_lables):
                txt += '\n'
                txt += '@@@ Operation {0}: {1}'.format(i + 1, group)
                txt += '\nSymmetry Transformation matrix\n'
                for array in sym_matrices[i]:
                    txt += ' {:11.6f} {:11.6f} {:11.6f}\n'.format(array[0], array[1], array[2])
                txt += '\n'
                txt += 'Symmetry Transformed Atomic Coordinates (Angstroms)\n'

                for idn, array in enumerate(geometry.get_positions()):
                    array2 = np.dot(array, sym_matrices[i].T)
                    txt += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idn],
                                                                      array2[0], array2[1], array2[2])
        output.write(txt)

    def print_wnfsym_sym_ovelap(self, group, axis=None, axis2=None, center=None, output=sys.stdout):

        txt = ''
        for molecule in self._molecules:
            mo_overlap = molecule.get_mo_overlaps(group, axis=axis, axis2=axis2, center=center)
            wf_overlap = molecule.get_wf_overlaps(group, axis=axis, axis2=axis2, center=center)

            ideal_gt = molecule.get_ideal_group_table(group, axis=axis, axis2=axis2, center=center)
            wf_measure = molecule.get_wf_symmetry(group, axis=axis, axis2=axis2, center=center)

            sep_line = '     ' + '---------' * len(ideal_gt['ir_labels']) + '\n'

            txt += '\nMolecule : {}\n'.format(molecule.name)

            txt += '\nIdeal Group Table\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in ideal_gt['labels']])
            txt += '\n'
            txt += sep_line
            for i, line in enumerate(ideal_gt['table']):
                txt += '{:4}'.format(ideal_gt['ir_labels'][i]) + '  '.join(['{:7.3f}'.format(s) for s in line])
                txt += '\n'
            txt += sep_line

            txt += '\nAlpha MOs: Symmetry Overlap Expectation Values\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in ideal_gt['labels']])
            txt += '\n'
            txt += sep_line

            for i, line in enumerate(mo_overlap['alpha']):
                txt += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
                txt += '\n'

            txt += '\nBeta MOs: Symmetry Overlap Expectation Values\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in ideal_gt['labels']])
            txt += '\n'
            txt += sep_line
            for i, line in enumerate(mo_overlap['beta']):
                txt += '{:4d}'.format(i + 1) + '  '.join(['{:7.3f}'.format(s) for s in line])
                txt += '\n'

            txt += '\nWaveFunction: Symmetry Overlap Expectation Values\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in ideal_gt['labels']])
            txt += '\n'
            txt += sep_line
            txt += 'a-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_overlap['alpha']])
            txt += '\n'
            txt += 'b-wf' + '  '.join(['{:7.3f}'.format(s) for s in wf_overlap['beta']])
            txt += '\n'
            txt += 'WFN ' + '  '.join(['{:7.3f}'.format(s) for s in wf_overlap['total']])
            txt += '\n'

            txt += '\nWaveFunction: CSM-like values\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in wf_measure['labels']])
            txt += '\n'
            txt += sep_line

            txt += 'Grim' + '  '.join(['{:7.3f}'.format(s) for s in wf_measure['grim']])
            txt += '\n'
            txt += 'CSM ' + '  '.join(['{:7.3f}'.format(s) for s in wf_measure['csm']])
            txt += '\n'

        output.write(txt)

    def print_wnfsym_irreducible_repr(self, group, axis=None, axis2=None, center=None, show_axes=True,
                                      output=sys.stdout):

        txt = ''
        for molecule in self._molecules:
            ir_mo = molecule.get_mo_irreducible_representations(group, axis=axis, axis2=axis2, center=center)
            data_wf = molecule.get_wf_irreducible_representations(group, axis=axis, axis2=axis2, center=center)

            txt += '\nMolecule : {}\n'.format(molecule.name)
            sep_line = '     ' + '---------' * (len(ir_mo['labels'])) + '\n'

            txt += '\nWaveFunction: Irred. Rep. Decomposition\n'
            txt += sep_line
            txt += '     ' + '  '.join(['{:^7}'.format(s) for s in ir_mo['labels']])
            txt += '\n'
            txt += sep_line
            txt += 'a-wf' + '  '.join(['{:7.3f}'.format(s) for s in data_wf['alpha']])
            txt += '\n'
            txt += 'b-wf' + '  '.join(['{:7.3f}'.format(s) for s in data_wf['beta']])
            txt += '\n'
            txt += 'WFN ' + '  '.join(['{:7.3f}'.format(s) for s in data_wf['total']])
            txt += '\n'

            if show_axes:
                txt += _get_axis_info(molecule, group, axis, axis2, center)

        output.write(txt)

    def print_wnfsym_mo_irreducible_repr(self, group, axis=None, axis2=None, center=None, show_axes=True,
                                         output=sys.stdout):

        txt = ''
        for molecule in self._molecules:
            ir_mo = molecule.get_mo_irreducible_representations(group, axis=axis, axis2=axis2, center=center)
            alpha_energies = molecule.electronic_structure.alpha_energies
            alpha_occupancy = molecule.electronic_structure.alpha_occupancy
            beta_energies = molecule.electronic_structure.beta_energies
            beta_occupancy = molecule.electronic_structure.beta_occupancy

            txt += '\nMolecule : {}\n'.format(molecule.name)

            sep_line = '     ' + '---------' * (2 + len(ir_mo['labels'])) + '\n'

            txt += '\nAlpha MOs: Irred. Rep. Decomposition\n'
            txt += sep_line
            txt += '     ' + '{:^10}'.format('occup') + '{:^8}'.format('E(eV)') + \
                   '  '.join(['{:^7}'.format(s) for s in ir_mo['labels']])
            txt += '\n'
            txt += sep_line
            non_one = False
            for i, line in enumerate(ir_mo['alpha']):
                txt += '{:4d}'.format(i + 1) + '{:6d}'.format(alpha_occupancy[i]) + \
                       '{:10.2f}'.format(alpha_energies[i]) + ' '*2 + '  '.join(['{:7.3f}'.format(s) for s in line])
                if abs(sum(line) - 1) > 1e-2:
                    txt += '*'
                    non_one = True
                txt += '\n'

            txt += '\nBeta MOs: Irred. Rep. Decomposition\n'
            txt += sep_line
            txt += '     ' + '{:^10}'.format('occup') + '{:^8}'.format('E(eV)') + \
                   '  '.join(['{:^7}'.format(s) for s in ir_mo['labels']])
            txt += '\n'
            txt += sep_line
            for i, line in enumerate(ir_mo['beta']):
                txt += '{:4d}'.format(i + 1) + '{:6d}'.format(beta_occupancy[i]) + \
                       '{:10.2f}'.format(beta_energies[i]) + ' '*2 + '  '.join(['{:7.3f}'.format(s) for s in line])
                if abs(sum(line) - 1) > 1e-2:
                    txt += '*'
                    non_one = True
                txt += '\n'

            if non_one:
                warn('The sum of some molecular orbital irreducible represaentations are not equal one (*)')

            if show_axes:
                txt += _get_axis_info(molecule, group, axis, axis2, center)

        output.write(txt)

    def plot_mo_diagram(self, group, axis=None, axis2=None, center=None):
        """
        Plots Molecular orbitals diagram

        :param group:
        :param axis:
        :param axis2:
        :param center:
        :return:
        """

        for molecule in self._molecules:

            ir_mo = molecule.get_mo_irreducible_representations(group, axis=axis, axis2=axis2, center=center)

            ird_a_max = [np.argmax(ird_a_orb) for ird_a_orb in ir_mo['alpha']]
            energies = molecule.electronic_structure.alpha_energies

            plt.figure()
            plt.title('{}'.format(molecule.name))

            ax1 = plt.axes()
            ax1.axes.get_xaxis().set_visible(False)  # Hide x axis
            # ax1.axes.get_yaxis().set_visible(True)

            degeneracy = [[energies[0]]]
            for energy in energies[1:]:
                if abs(energy - degeneracy[-1][-1]) < 1e-3:
                    degeneracy[-1].append(energy)
                else:
                    degeneracy.append([energy])

            max_value = 5e-3
            x_center = []
            for ix in degeneracy:
                if len(ix) == 1:
                    x_center.append([0])
                else:
                    x_center.append(np.linspace(-max_value, max_value, len(ix)))
            x_center = [y for x in x_center for y in x]

            plt.scatter(x_center, energies, s=500, marker="_", linewidth=3)
            for i in range(len(energies)):
                plt.text(-max_value * 2, energies[i], ir_mo['labels'][ird_a_max[i]])

        plt.show()

    def plot_sym_energy_evolution(self, group, axis=None, axis2=None, center=None):

        energies = []
        ird_a_max = []
        for idm, molecule in enumerate(self._molecules):

            ir_mo = molecule.get_mo_irreducible_representations(group, axis=axis, axis2=axis2, center=center)

            labels = ir_mo['labels']

            ird_a_max.append(np.array([np.argmax(ird_a_orb) for ird_a_orb in ir_mo['alpha']]))
            energies.append(molecule.electronic_structure.alpha_energies)

        energies_x_orbital = np.array(energies).T
        ird_a_x_orbital = np.array(ird_a_max).T

        for i in range(len(ird_a_x_orbital)):
            for j in range(len(ird_a_x_orbital[i])):
                if j == 0:
                    old_ird = ird_a_x_orbital[i][0]
                else:
                    if old_ird != ird_a_x_orbital[i][j]:
                        for k in range(len(ird_a_x_orbital) - i):
                            if old_ird == ird_a_x_orbital[k + i][j]:
                                ird_a_x_orbital[i], ird_a_x_orbital[k + i] = swap_vectors(ird_a_x_orbital[i],
                                                                                          ird_a_x_orbital[k + i], j)
                                energies_x_orbital[i], energies_x_orbital[k + i] = swap_vectors(energies_x_orbital[i],
                                                                                                energies_x_orbital[
                                                                                                    k + i],
                                                                                                j)
                                break
                old_ird = ird_a_x_orbital[i][j]

        for ide, energy in enumerate(energies_x_orbital):
            x = np.arange(len(energy))
            plt.plot(x, energy, marker='_')
            for i in range(len(energy)):
                plt.text(x[i], energy[i] + abs(energy[i]) * 0.001, labels[ird_a_x_orbital[ide][i]])

        plt.show()

    def get_shape_measure(self, label, kind, central_atom=0, fix_permutation=False):
        """
        Get shape measure

        :param label: Reference shape label
        :type label: str
        :param kind: function name suffix
        :type kind: str
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param fix_permutation: Do not permute atoms during shape calculations
        :type fix_permutation: bool
        :return: Shape measures
        :rtype: list
        """
        get_measure = 'get_shape_' + kind

        return [getattr(geometry, get_measure)(label, central_atom=central_atom, fix_permutation=fix_permutation)
                for geometry in self.get_geometries()]

    def get_molecule_path_deviation(self, shape_label1, shape_label2, central_atom=0):
        """
        Get molecule path deviation

        :param shape_label1: First shape reference label
        :type shape_label1: str
        :param shape_label2: Second shape reference label
        :type shape_label2: str
        :param central_atom: Position of the central atom
        :type central_atom: int
        """
        return [geometry.get_path_deviation(shape_label1, shape_label2, central_atom) for geometry
                in self.get_geometries()]

    def get_molecule_generalized_coord(self, shape_label1, shape_label2, central_atom=0):

        return [geometry.get_generalized_coordinate(shape_label1, shape_label2, central_atom)
                for geometry in self.get_geometries()]

    # TODO: This may be placed inside Shape class
    def get_path_parameters(self, shape_label1, shape_label2, central_atom=0):

        if type(shape_label1) is Geometry:
            label1 = shape_label1
            label1_name = shape_label1.name
        else:
            label1 = shape_label1
            label1_name = shape_label1
        if type(shape_label2) is Geometry:
            label2 = shape_label2
            label2_name = shape_label2.name
        else:
            label2 = shape_label2
            label2_name = shape_label2

        csm = {label1_name: self.get_shape_measure(label1, 'measure', central_atom),
               label2_name: self.get_shape_measure(label2, 'measure', central_atom)}
        devpath = self.get_molecule_path_deviation(label1, label2, central_atom)
        generalized_coord = self.get_molecule_generalized_coord(label1, label2, central_atom)

        return csm, devpath, generalized_coord

    def print_minimum_distortion_path_shape(self, shape_label1, shape_label2, central_atom=0,
                                            min_dev=0, max_dev=15, min_gco=0, max_gco=101,
                                            num_points=20, output=None):
        """
        Print the minimum distortion path

        :param shape_label1: First reference shape label
        :type shape_label1: str
        :param shape_label2: Second reference shape label
        :type shape_label2: str
        :param central_atom: Position of the central atom
        :type central_atom: int
        :param min_dev:
        :type min_dev: float
        :param max_dev:
        :type max_dev: float
        :param min_gco:
        :type min_gco: float
        :param max_gco:
        :type max_gco: float
        :param num_points: Number of points
        :type num_points: int
        :param output1: Display hook
        :type output1: hook
        """

        if output is not None:
            output_name, file_extension = os.path.splitext(output.name)
            output1 = open(output_name + '_pth.csv', 'w')
            output2 = open(output_name + '_pth.xyz', 'w')
            output3 = output
        else:
            output1 = sys.stdout
            output2 = sys.stdout
            output3 = sys.stdout

        csm, devpath, gen_coord = self.get_path_parameters(shape_label1, shape_label2, central_atom=central_atom)

        txt_params = 'Deviation threshold to calculate Path deviation function: ' \
                     '{:2.1f}% - {:2.1f}%\n'.format(min_dev, max_dev)
        txt_params += 'Deviation threshold to calculate Generalized Coordinate: ' \
                      '{:2.1f}% - {:2.1f}%\n'.format(min_gco, max_gco)
        txt_params += '\n'
        txt_params += '{:9} '.format('structure'.upper())
        for csm_label in list(csm.keys()):
            txt_params += '{:^8} '.format(csm_label)
        txt_params += '{:^8} {:^8}'.format('DevPath', 'GenCoord')
        txt_params += '\n'

        filter_mask = [min_dev < dv < max_dev and min_gco < gc < max_gco for dv, gc in zip(devpath, gen_coord)]

        for idx, molecule in enumerate(self._molecules):
            if not filter_mask[idx]:
                continue

            txt_params += '{:9} '.format(molecule.name.strip())
            for label in list(csm.keys()):
                txt_params += '{:^8.3f} '.format(csm[label][idx])
            txt_params += '{:^8.1f} {:^8.1f}'.format(devpath[idx], gen_coord[idx])
            txt_params += '\n'

        # self.print_shape_measure()
        txt_params += 'skipped {} structure/s\n\n'.format(filter_mask.count(False))
        output3.write(txt_params)

        if isinstance(shape_label1, Geometry):
            label1_name = shape_label1.name
        else:
            label1_name = shape_label1
        if isinstance(shape_label2, Geometry):
            label2_name = shape_label2.name
        else:
            label2_name = shape_label2

        path = get_shape_path(shape_label1, shape_label2, num_points)
        txt_path = 'Minimum distortion path\n'
        txt_path += ' {:^6}  {:^6}\n'.format(label1_name, label2_name)
        for idx, value in enumerate(path[0]):
            txt_path += '{:6.3f}  {:6.3f}'.format(path[0][idx], path[1][idx])
            txt_path += '\n'
        txt_path += '\n'
        output1.write(txt_path)

        test_structures = []
        for ids, structure in enumerate(path[2]):
            test_structures.append(Geometry(symbols=['' for _ in range(len(structure))],
                                            positions=structure, name='map_structure{}'.format(ids)))
        output2.write(file_io.get_file_xyz_txt(test_structures))

        if output1 is sys.stdout:
            import matplotlib.pyplot as plt
            plt.plot(path[0], path[1], 'k', linewidth=2.0)
            plt.scatter(np.array(csm[label1_name])[filter_mask],
                        np.array(csm[label2_name])[filter_mask], linewidths=0.01)
            plt.xlabel(label1_name)
            plt.ylabel(label2_name)
            plt.show()

    def get_point_group(self, tol=0.01):
        """
        Get the point group of all structures

        :param tol: Tolerance
        :type tol: float
        :return: a list of point group labels
        :rtype: list
        """
        return [molecule.geometry.get_pointgroup(tol) for molecule in self._molecules]



import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return ' %s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)


warnings.formatwarning = warning_on_one_line


class MissingAtomWarning(Warning):
    pass


class DifferentVerteciesWarning(Warning):
    pass


class EmptyLineWarning(Warning):
    pass



import yaml, sys
from cosymlib.molecule.geometry import Geometry
from cosymlib import __version__


def print_header(output=sys.stdout):
    output.write('-' * 70 + '\n')
    output.write(' COSYMLIB v{}\n Electronic Structure & Symmetry Group\n'.format(__version__))
    output.write(' Institut de Quimica Teorica i Computacional (IQTC)\n')
    output.write(' Universitat de Barcelona\n')
    output.write('-' * 70 + '\n\n')


def print_footer(output=sys.stdout):
    output.write('\n' + '-' * 70 + '\n')
    output.write(' ' * 20 + 'End of calculation\n')
    output.write('-' * 70 + '\n\n')


def print_input_info(initial_geometries, output=sys.stdout):

    for ids, geometry in enumerate(initial_geometries):
        output.write('Structure {} : {}\n'.format(ids+1, geometry.name))

        for idn, array in enumerate(geometry.get_positions()):
            output.write('{:2s}'.format(geometry.get_symbols()[idn]))
            output.write(' {:11.7f} {:11.7f} {:11.7f}'.format(array[0], array[1], array[2]))

            if geometry.get_connectivity() is not None:
                output.write('  ' +
                             '  '.join(['{:3}'.format(i+1) for i in range(geometry.get_n_atoms())
                                        if [idn+1, i+1] in geometry.get_connectivity()])+
                             '\n')
        output.write('\n')


def add_extra_keywords(args, filename):
    with open(filename, 'r') as stream:
        input_parameters = yaml.load(stream, Loader=yaml.FullLoader)
    try:
        for key, value in input_parameters.items():
            if key.lower() in args:
                setattr(args, key.lower(), value)
            else:
                raise KeyError("'{}' is not a valid input keyword".format(key))
    except AttributeError:
        raise KeyError('Incorrect input file format')


def extract_geometries(structure_list, as_list=False):
    """
    extracts the Geometry from Molecule or list of Molecule
    if Geometry or list of Geometry is introduced no changes are made

    :param structure_list:
    :param as_list:
    :return:
    """

    if isinstance(structure_list, list):
        geometry_list = []
        n_atoms_ref = None
        for structure in structure_list:
            if isinstance(structure, Geometry):
                geometry_list.append(structure)
                n_atoms = structure.get_n_atoms()
            else:
                geometry_list.append(structure.geometry)
                n_atoms = structure.geometry.get_n_atoms()

            if n_atoms_ref is not None:
                if n_atoms != n_atoms_ref:
                    raise Exception('All geometries should contain the same number of atoms!')
                n_atoms_ref = structure.get_n_atoms()

        return geometry_list

    else:
        if isinstance(structure_list, Geometry):
            return [structure_list] if as_list else structure_list
        else:
            return [structure_list.geometry] if as_list else structure_list.geometry




import os
import yaml
import numpy as np
from warnings import warn


file_path = os.path.dirname(os.path.abspath(__file__)) + '/periodic_table.yaml'

with open(file_path, 'r') as stream:
    periodic_table_info = yaml.safe_load(stream)


def atomic_number_to_element(z):
    for element, data in periodic_table_info.items():
        if int(z) == data['atomic_number']:
            return element


def element_to_atomic_number(symbol):
    return periodic_table_info[symbol.capitalize()]['atomic_number']


def element_mass(symbol):
    return periodic_table_info[symbol.capitalize()]['atomic_mass']


def element_valence_electron(symbol):
    return periodic_table_info[symbol.capitalize()]['valence_electrons']


def center_mass(elements, coordinates):
    cm = [0., 0., 0.]
    m = 0
    for ide, element in enumerate(elements):
        cm += coordinates[ide] * element_mass(element)
        m += element_mass(element)
    cm = cm/m
    return cm


def generate_connectivity_from_geometry_slow(geometry, thresh=1.2):
    coor = geometry.get_positions()
    sym = geometry.get_symbols()

    connectivity = []
    for i, (atom1, sym1) in enumerate(zip(np.array(coor), sym)):
        for j, (atom2, sym2) in enumerate(zip(np.array(coor), sym)):
            dist = np.linalg.norm(atom1 - atom2)
            rad = periodic_table_info[sym1]['covalent_radius'] + periodic_table_info[sym2]['covalent_radius']

            if np.abs(rad - dist)/rad < thresh - 1:
                connectivity.append((i+1, j+1))

    return connectivity


def generate_connectivity_from_geometry(geometry, thresh=1.2):
    from scipy.spatial import distance_matrix
    coordinates = geometry.get_positions()
    try:
        radius = [periodic_table_info[sym]['covalent_radius'] for sym in geometry.get_symbols()]
    except KeyError:
        warn('failed to generate connectivity, no connectivity will be used')
        return None

    distances_matrix = distance_matrix(coordinates, coordinates)

    radii_matrix = np.array([radius]*len(radius))
    radii_matrix = radii_matrix + radii_matrix.T

    try:
        relative_differences = np.abs(radii_matrix - distances_matrix)/radii_matrix
    except ValueError:
        warn('failed to generate connectivity')
        return None

    return (np.array(np.where(relative_differences < thresh - 1)).T + 1).tolist()


def get_connectivity_matrix(connectivity, ndim):
    cm = np.zeros((ndim, ndim), dtype=int)
    for pair in connectivity:
        cm[pair[0]-1, pair[1]-1] = 1
    return cm



from cosymlib.file_io import get_geometry_from_file_cor
from cosymlib.file_io import errors

import os
import tempfile
import warnings


def read_old_input(file_name):
    """
    Reads the old Shape's program input
    :param file_name: file name
    :return: list of Geometry objects and options
    """

    options = {'%out': None, '%conquest': None, '%external': False, '%fullout': False, '%test': False,
               '%n_atoms': 0, '%central_atom': 0, '%labels': 0, '%path': False}

    idl = 0
    with open(file_name, mode='r') as lines:
        while True:
            line = lines.readline().split()
            if '$' in line or '!' in line:
                pass
            elif any('%' in word for word in line):
                if len(line) > 1:
                    options[line[0]] = line[1]
                else:

                    options[line[0]] = True
            else:
                try:
                    int(line[0])
                    if options['%n_atoms'] == 0:
                        options['%n_atoms'] = int(line[0])
                        options['%central_atom'] = int(line[1])
                    else:
                        options['%labels'] = line
                except (ValueError, IndexError):
                    break
            idl += 1

        n_atoms = options['%n_atoms']
        if options['%central_atom'] != 0:
            n_atoms += 1
        if options['%conquest'] is not None:
            dir = os.path.dirname(os.path.abspath(file_name))
            structures = get_geometry_from_file_cor(os.path.join(dir, options['%conquest'] + '.cor'), read_multiple=True)
        else:
            tmp = tempfile.NamedTemporaryFile(mode='w+t', dir=os.getcwd())
            tmp_lines = lines.readlines()
            try:
                # Write data to the temporary file
                tmp.write(line[0]+'\n')
                idl += 1
                for line in tmp_lines:
                    if line.strip() == '':
                        warnings.warn('Line {} is empty'.format(idl + 1), errors.EmptyLineWarning)
                    else:
                        tmp.write(line)
                    idl += 1
                tmp.seek(0)
                structures = get_geometry_from_file_cor(tmp.name, read_multiple=True)
            finally:
                tmp.close()

    return [structures, options]



from cosymlib.molecule import Molecule
from cosymlib.molecule.geometry import Geometry
from cosymlib.molecule.electronic_structure import ElectronicStructure
from cosymlib.file_io.tools import extract_geometries
from cosymlib.tools import atomic_number_to_element, element_to_atomic_number
from cosymlib.file_io import errors

import numpy as np
import os, re
import warnings


comment_line = ('#', '!', '$')


def _non_blank_lines(f):
    for l in f:
        line = l.rstrip()
        if line:
            yield line


def check_geometries_vertices(geometries, file_name):
    n_atoms = geometries[0].get_n_atoms()
    for idg, geometry in enumerate(geometries):
        if geometry.get_n_atoms() != n_atoms:
            warnings.warn('Structure {} with name {} of file {}\n '
                          'Number of vertices does not match with other structures\n'.format(idg + 1, geometry.name,
                                                                                             file_name),
                          errors.DifferentVerteciesWarning)


# Read INPUT files
def read_generic_structure_file(input_name, read_multiple=False):
    # print('Reading file {}...'.format(os.path.basename(input_name)))
    if os.stat(input_name).st_size == 0:
        raise FileExistsError('File {} is empty'.format(os.path.basename(input_name)))
    file_name, file_extension = os.path.splitext(input_name)
    if 'molden' in file_name:
        file_extension = ' molden'

    possibles = globals().copy()
    possibles.update(locals())
    method = possibles.get('get_molecule_from_file_' + file_extension[1:])
    if method is None:
        method = possibles.get('get_geometry_from_file_' + file_extension[1:])
    if not method:
        raise NotImplementedError("File not recognized")

    return method(input_name, read_multiple=read_multiple)


def get_geometry_from_file_xyz(file_name, read_multiple=False):
    """
    Reads a XYZ file and returns the geometry of all structures in it
    :param file_name: file name
    :param read_multiple: read multiple files if available
    :return: list of Geometry objects
    """
    input_molecule = [[], []]
    geometries = []
    n_atoms = None
    n_atoms_line = 1
    with open(file_name, mode='r') as lines:
        for idl, line in enumerate(lines):
            if line.lstrip().startswith(comment_line):
                pass
            elif line.strip() == '':
                warnings.warn('Line {} is empty'.format(idl + 1), errors.EmptyLineWarning)
            else:
                try:
                    float(line.split()[1])
                    input_molecule[0].append(line.split()[0])
                    input_molecule[1].append(line.split()[1:])
                except (IndexError, ValueError):
                    if input_molecule[0]:
                        if len(input_molecule[0]) != n_atoms:
                            warnings.warn('Number of atoms in line {} and number '
                                          'of atoms provided are not equal'.format(n_atoms_line),
                                          errors.MissingAtomWarning)
                        geometries.append(Geometry(symbols=input_molecule[0],
                                                   positions=input_molecule[1],
                                                   name=name))
                        n_atoms = None

                    input_molecule = [[], []]
                    if n_atoms is None:
                        n_atoms = int(line.split()[0])
                        n_atoms_line = idl + 1
                    else:
                        name = line.split()[0]

        if len(input_molecule[0]) != n_atoms:
            warnings.warn('Number of atoms in line {} and number '
                          'of atoms provided are not equal'.format(n_atoms_line),
                          errors.MissingAtomWarning)

        geometries.append(Geometry(symbols=input_molecule[0],
                                   positions=input_molecule[1],
                                   name=name))
        if not read_multiple:
            return geometries[0]

    # Check if all geometries are of the same vertices
    check_geometries_vertices(geometries, file_name)

    return geometries


def get_geometry_from_file_cor(file_name, read_multiple=False):
    """
    Reads a Conquest formatted file and the geometry of all structures in it
    :param file_name: file name
    :param read_multiple: read multiple files if available
    :return: list of Geometry objects
    """
    input_molecule = [[], []]
    geometries = []
    with open(file_name, mode='r') as lines:

        for idl, line in enumerate(lines):
            if line.lstrip().startswith(comment_line):
                pass
            elif line.strip() == '':
                warnings.warn('Line {} is empty'.format(idl + 1), errors.EmptyLineWarning)
            else:
                try:
                    float(line.split()[1])
                    if len(line.split()) == 4:
                        input_molecule[0].append(line.split()[0])
                        input_molecule[1].append(line.split()[1:])
                    elif len(line.split()) == 5:
                        input_molecule[0].append(line.split()[0])
                        input_molecule[1].append(line.split()[1:-1])
                    else:
                        float(line.split()[2])
                except (ValueError, IndexError):
                    if input_molecule[0]:
                        geometries.append(Geometry(symbols=input_molecule[0],
                                                   positions=input_molecule[1],
                                                   name=name))
                    input_molecule = [[], []]
                    name = line.replace('**FRAG**', '')[:-1]
                    name = name.replace(' ', '')
        geometries.append(Geometry(symbols=input_molecule[0],
                                   positions=input_molecule[1],
                                   name=name))
        if not read_multiple:
            return geometries[0]

    # Check if all geometries are of the same vertices
    check_geometries_vertices(geometries, file_name)

    return geometries


# TODO: This function should handle open and close shell, mo energies and more. A lot to improve!
def get_molecule_from_file_fchk(file_name, read_multiple=False):
    key_list = ['Charge', 'Multiplicity', 'Number of alpha electrons', 'Number of beta electrons', 'Atomic numbers',
                'Current cartesian coordinates', 'Shell type', 'Number of primitives per shell', 'Shell to atom map',
                'Primitive exponents', 'Contraction coefficients', 'P(S=P) Contraction coefficients',
                'Alpha Orbital Energies', 'Beta Orbital Energies', 'Alpha MO coefficients', 'Beta MO coefficients']
    input_molecule = [[] for _ in range(len(key_list))]
    read = False
    with open(file_name, mode='r') as lines:
        name = lines.readline().split()[0]#.strip('\n')
        line = lines.readline().split()
        basis_set = line[-1]
        if 'R' in line[1]:
            del key_list[-1]

        n = 1
        options = True
        for line in lines:
            if read:
                try:
                    float(line.split()[0])
                    input_molecule[n].append(line.split())
                except ValueError:
                    read = False

            for idn, key in enumerate(key_list):
                if key in line:
                    if n == len(key_list) - 1:
                        break
                    if options and idn != 4:
                        input_molecule[idn].append(int(line.split()[-1]))
                        n = idn
                        break
                    else:
                        options = False
                    if n == idn:
                        n += 1
                    else:
                        n = idn
                    read = True
                    break

        for n in range(4, len(input_molecule)):
            input_molecule[n] = reformat_input(input_molecule[n])
        bohr_to_angstrom = 0.529177249
        coordinates = np.array(input_molecule[5], dtype=float).reshape(-1, 3) * bohr_to_angstrom

        atomic_number = [int(num) for num in input_molecule[4]]
        symbols = []
        for number in atomic_number:
            symbols.append(atomic_number_to_element(number))

        basis = basis_format(basis_set_name=basis_set,
                             atomic_numbers=atomic_number,
                             atomic_symbols=symbols,
                             shell_type=input_molecule[6],
                             n_primitives=input_molecule[7],
                             atom_map=input_molecule[8],
                             p_exponents=input_molecule[9],
                             c_coefficients=input_molecule[10],
                             p_c_coefficients=input_molecule[11])

        geometry = Geometry(symbols=symbols,
                            positions=coordinates,
                            name=name)

        alpha_energies = np.array(input_molecule[12], dtype=float).tolist()
        if input_molecule[13]:
            beta_energies = np.array(input_molecule[13], dtype=float).tolist()
        else:
            beta_energies = []

        Ca = np.array(input_molecule[14], dtype=float).reshape(-1, int(np.sqrt(len(input_molecule[14]))))
        if input_molecule[15]:
            Cb = np.array(input_molecule[15], dtype=float).reshape(-1, int(np.sqrt(len(input_molecule[15]))))
        else:
            Cb = []

        electronic_structure = ElectronicStructure(basis=basis,
                                                   orbital_coefficients=[Ca, Cb],
                                                   # charge=input_molecule[0][0],
                                                   multiplicity=input_molecule[1][0],
                                                   alpha_energies=alpha_energies,
                                                   beta_energies=beta_energies,
                                                   alpha_occupancy=[1]*int(input_molecule[2][0]),
                                                   beta_occupancy=[1]*int(input_molecule[3][0]))

        if read_multiple:
            return [Molecule(geometry, electronic_structure)]

        return Molecule(geometry, electronic_structure)


def get_molecule_from_file_molden(file_name, read_multiple=False):
    key_list = ['Charge', 'Multiplicity', 'Atomic numbers', 'Current cartesian coordinates',
                'Shell type', 'Number of primitives per shell', 'Shell to atom map', 'Primitive exponents',
                'Contraction coefficients', 'P(S=P) Contraction coefficients',
                'Alpha MO coefficients', 'Beta MO coefficients', 'MO Energies']

    type_list = {'s': '0',
                 'p': '1',
                 'd': '2',
                 'f': '3',
                 'sp': '-1'}

    input_molecule = {key: [] for key in key_list}
    read_molden = False
    read_coordinates = False
    read_basis = False
    read_coefficients = False
    occupation = {'Alpha': [], 'Beta': []}
    with open(file_name, mode='r') as lines:
        lines.readline()
        lines.readline()
        name = file_name.split('.')[0]
        for line in _non_blank_lines(lines):

            if '[' in line:
                pass
            elif read_molden:
                if '======= END OF MOLDEN-FORMAT MOLECULAR ORBITALS =======' in line:
                    break
                if read_coordinates:
                    input_molecule['Atomic numbers'].append(line.split()[2])
                    input_molecule['Current cartesian coordinates'].append(line.split()[3:])
                if read_basis:
                    try:
                        number = float(line.split()[0].replace('D', 'E'))
                        if number-int(number) != 0.:
                            input_molecule['Primitive exponents'].append(line.split()[0].replace('D', 'E'))
                            input_molecule['Contraction coefficients'].append(line.split()[1].replace('D', 'E'))
                            if len(line.split()) > 2:
                                input_molecule['P(S=P) Contraction coefficients'].append(line.split()[2].replace('D', 'E'))
                            else:
                                input_molecule['P(S=P) Contraction coefficients'].append(0.)
                        else:
                            atom = line.split()[0]
                    except ValueError:
                        input_molecule['Shell to atom map'].append(atom)
                        input_molecule['Shell type'].append(type_list[line.split()[0].lower()])
                        input_molecule['Number of primitives per shell'].append(line.split()[1])
                if read_coefficients:
                    if 'Sym' in line:
                        pass
                    else:
                        if 'Ene' in line:
                            input_molecule['MO Energies'].append(float(re.split('[= ]', line)[-1]))
                        elif 'Spin' in line:
                            spin = re.split('[= ]', line)[-1]
                            input_molecule[spin + ' MO coefficients'].append([])
                        elif 'Occup' in line:
                            occupation[spin].append(int(float(line.split('=')[-1])))
                        else:
                            input_molecule[spin+' MO coefficients'][-1].append(line.split()[1])

            if '[Atoms]' in line:
                read_molden = True
                read_coordinates = True
                if 'AU' in line:
                    bohr_to_angstrom = 0.529177249
                else:
                    bohr_to_angstrom = 1
            elif '[GTO]' in line:
                read_coordinates = False
                read_basis = True
            elif '[MO]' in line:
                read_basis = False
                read_coefficients = True
            elif '[5D]' in line:
                input_molecule['Shell type'] = [x.replace('2', '-2') for x in input_molecule['Shell type']]
            elif '[7F]' in line:
                input_molecule['Shell type'] = [x.replace('3', '-3') for x in input_molecule['Shell type']]

        coordinates = np.array(input_molecule['Current cartesian coordinates'], dtype=float).reshape(-1, 3) * bohr_to_angstrom

        total_n_electrons = 0
        symbols = []
        for atom_number in input_molecule['Atomic numbers']:
            total_n_electrons += int(atom_number)
            symbols.append(atomic_number_to_element(atom_number))

        new_occupation = []
        if len(occupation['Beta']) == 0:
            if 2 not in occupation['Alpha']:
                occupation['Beta'] = occupation['Alpha']
            else:
                for occup in occupation['Alpha']:
                    new_occupation.append(occup//2)
                    if occup == 2:
                        occupation['Beta'].append(1)
                    else:
                        occupation['Beta'].append(0)
                occupation['Alpha'] = new_occupation
        n_electrons = sum(occupation['Alpha'] + occupation['Beta'])
        input_molecule['Multiplicity'].append(sum(occupation['Alpha']) - sum(occupation['Beta']) + 1)
        input_molecule['Charge'].append(int(total_n_electrons - n_electrons))

        basis = basis_format(basis_set_name='UNKNOWN',
                             atomic_numbers=input_molecule['Atomic numbers'],
                             atomic_symbols=symbols,
                             shell_type=input_molecule['Shell type'],
                             n_primitives=input_molecule['Number of primitives per shell'],
                             atom_map=input_molecule['Shell to atom map'],
                             p_exponents=input_molecule['Primitive exponents'],
                             c_coefficients=input_molecule['Contraction coefficients'],
                             p_c_coefficients=input_molecule['P(S=P) Contraction coefficients'])

        geometry = Geometry(symbols=symbols,
                            positions=coordinates,
                            name=name)

        Ca = np.array(input_molecule['Alpha MO coefficients'], dtype=float)
        if input_molecule['Beta MO coefficients']:
            Cb = np.array(input_molecule['Beta MO coefficients'], dtype=float)
        else:
            Cb = []

        warnings.warn('Beta energies not implemented yet')
        ee = ElectronicStructure(basis=basis,
                                 orbital_coefficients=[Ca, Cb],
                                 # charge=input_molecule['Charge'][0],
                                 multiplicity=input_molecule['Multiplicity'][0],
                                 alpha_energies=input_molecule['MO Energies'],
                                 beta_energies=[],
                                 alpha_occupancy=occupation['Alpha'],
                                 beta_occupancy=occupation['Beta'])

        if read_multiple:
            return [Molecule(geometry, ee)]

        return Molecule(geometry, ee)


def get_geometry_from_file_ref(file_name, read_multiple=False):
    """
    Reads a Conquest formatted file and the geometry of all structures in it
    :param file_name: file name
    :param read_multiple: read all geometries inside the file
    :return: list of Geometry objects
    """
    input_molecule = []
    structures = []
    with open(file_name, mode='r') as lines:
        lines.readline()
        lines.readline()
        name = lines.readline().split()[0]
        for line in lines:
            if line.lstrip().startswith(comment_line):
                pass
            # elif line.strip() == '':
            #     warnings.warn('Line {} is empty'.format(idl + 1), custom_errors.EmptyLineWarning)
            else:
                try:
                    if len(line.split()) > 3:
                        float(line.split()[1])
                        input_molecule.append(line.split()[1:])
                    else:
                        float(line.split()[0])
                        input_molecule.append(line.split())
                except (ValueError, IndexError):
                    if input_molecule:
                        structures.append(Geometry(positions=input_molecule,
                                                   name=name))
                    input_molecule = []
                    name = line.split()[0]
        structures.append(Geometry(positions=input_molecule,
                                   name=name))
    if read_multiple is False:
        return structures[0]

    return structures


def get_geometry_from_file_pdb(file_name, read_multiple=False):
    """
    Reads a Conquest formatted file and the geometry of all structures in it
    :param file_name: file name
    :param read_multiple: read multiple files if available
    :return: list of Geometry objects
    """

    file_txt = open(file_name, mode='r').read()

    index_list = []
    for i in re.finditer('TITLE', file_txt):
        index_list.append(i.start())
    index_list.append(len(file_txt))

    geometries = []
    for i in range(len(index_list)-1):
        mol_section = file_txt[index_list[i]:index_list[i+1]]
        coordinates = []
        symbols = []
        connect = []
        name = ''
        for line in mol_section.split('\n'):
            if line.find('TITLE') > -1:
                name = ' '.join(line.split()[1:])
            if line.find('HETATM') > -1 or line.find('ATOM') > -1:
                coordinates.append([float(num) for num in line.split()[4:7]])
                symbols.append(line.split()[2])
            if line.find('CONECT') > -1:
                connect.append([int(num) for num in line.split()[1:]])
            if line.find('TER') > -1 or line.find('END') > -1:
                break

        connectivity = []
        for atom in connect:
            for j in atom[1:]:
                connectivity.append((atom[0], j))

        geometries.append(Geometry(symbols=symbols,
                                   positions=coordinates,
                                   name=name,
                                   connectivity=connectivity))

    if read_multiple is False:
        return geometries[0]

    return geometries


def basis_format(basis_set_name,
                 atomic_numbers,
                 atomic_symbols,
                 shell_type,
                 n_primitives,
                 atom_map,
                 p_exponents,
                 c_coefficients,
                 p_c_coefficients):
    type_list = {'0': ['s', 1],
                 '1': ['p', 3],
                 '2': ['d', 6],
                 '3': ['f', 10],
                 '-1': ['sp', 4],
                 '-2': ['d_', 5],
                 '-3': ['f_', 7]}

    atomic_numbers = np.array(atomic_numbers, dtype=int)
    atom_map = np.array(atom_map, dtype=int)

    basis_set = {'name': basis_set_name,
                 'primitive_type': 'gaussian'}

    # shell_type_index = [0] + np.cumsum([type_list['{}'.format(s)][1]
    #                                     for s in shell_type]).tolist()
    prim_from_shell_index = [0] + np.cumsum(np.array(n_primitives, dtype=int)).tolist()

    atoms_data = []
    for iatom, atomic_number in enumerate(atomic_numbers):
        symbol = atomic_symbols[iatom]

        shell_from_atom_counts = np.unique(atom_map, return_counts=True)[1]
        shell_from_atom_index = np.unique(atom_map, return_index=True)[1]

        shells_data = []
        for ishell in range(shell_from_atom_counts[iatom]):
            st = type_list['{}'.format(shell_type[shell_from_atom_index[iatom] + ishell])]
            ini_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell]
            fin_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell + 1]

            shells_data.append({
                'shell_type': st[0],
                'p_exponents': p_exponents[ini_prim: fin_prim],
                'con_coefficients': c_coefficients[ini_prim: fin_prim],
                'p_con_coefficients': p_c_coefficients[ini_prim: fin_prim],
            })

        atoms_data.append({'shells': shells_data,
                           'symbol': symbol,
                           'atomic_number': atomic_number})

    basis_set['atoms'] = atoms_data

    return basis_set


def get_connectivity_from_file(file_name):

    file_txt = open(file_name, mode='r').read()

    connect = []
    for line in file_txt.split('\n'):
            connect.append([int(num) for num in line.split()])

    connectivity = []
    for atom in connect:
        for j in atom[1:]:
            connectivity.append((atom[0], j))

    return connectivity


# Get OUPUT files
def get_file_xyz_txt(structure):

    # Get xyz file format from Molecule or Geometry (or list)
    geometries = extract_geometries(structure, as_list=True)

    txt = ''
    for geometry in geometries:
        txt += '{}\n'.format(geometry.get_n_atoms())
        txt += '{}\n'.format(geometry.name if geometry.name is not None else '')
        for idp, position in enumerate(geometry.get_positions()):
            txt += '{:2} {:11.6f} {:11.6f} {:11.6f}\n'.format(geometry.get_symbols()[idp],
                                                                  position[0], position[1], position[2])
    return txt


# Support functions
def reformat_input(array):
    flat_list = []
    for sublist in array:
        for item in sublist:
            if len(item) > 2:
                flat_list.append(item)
            else:
                flat_list.append(item)
    return flat_list


def get_array_txt(label, type, array, row_size=5):

    formats = {'R': '15.8e',
               'I': '11'}

    n_elements = len(array)
    rows = int(np.ceil(n_elements/row_size))

    txt_fchk = '{:40}   {}   N=       {:5}\n'.format(label, type, n_elements)

    for i in range(rows):
        if (i+1)*row_size > n_elements:
            txt_fchk += (' {:{fmt}}'* (n_elements - i*row_size) + '\n').format(*array[i * row_size:n_elements],
                                                                               fmt=formats[type])
        else:
            txt_fchk += (' {:{fmt}}'* row_size  + '\n').format(*array[i * row_size: (i+1)*row_size],
                                                               fmt=formats[type])

    return txt_fchk


def build_fchk(molecule):


    structure = np.array(molecule.geometry.get_positions())
    basis = molecule.electronic_structure.basis
    alpha_mo_coeff = molecule.electronic_structure.coefficients_a
    alpha_mo_energies = molecule.electronic_structure.alpha_energies
    if molecule.electronic_structure.coefficients_b != molecule.electronic_structure.coefficients_a:
        beta_mo_coeff = molecule.electronic_structure.coefficients_b
        beta_mo_energies = molecule.electronic_structure.beta_energies
        beta_mo_coeff = np.array(beta_mo_coeff).flatten().tolist()
    else:
        beta_mo_coeff = None
        beta_mo_energies = None

    number_of_basis_functions = len(alpha_mo_coeff)
    alpha_electrons = molecule.electronic_structure.alpha_electrons
    beta_electrons = molecule.electronic_structure.beta_electrons
    number_of_electrons = alpha_electrons + beta_electrons

    alpha_mo_coeff = np.array(alpha_mo_coeff).flatten().tolist()

    atomic_numbers = [element_to_atomic_number(symbol) for symbol in molecule.geometry.get_symbols()]
    total_energy = np.dot(alpha_mo_energies, alpha_electrons) + np.dot(beta_mo_energies, beta_electrons)

    shell_type_list = {'s':  {'type':  0, 'angular_momentum': 0},
                       'p':  {'type':  1, 'angular_momentum': 1},
                       'd':  {'type':  2, 'angular_momentum': 2},
                       'f':  {'type':  3, 'angular_momentum': 3},
                       'sp': {'type': -1, 'angular_momentum': 1},  # hybrid
                       'd_': {'type': -2, 'angular_momentum': 2},  # pure
                       'f_': {'type': -3, 'angular_momentum': 3}}  # pure

    shell_type = []
    p_exponents = []
    c_coefficients = []
    p_c_coefficients = []
    n_primitives = []
    atom_map = []

    largest_degree_of_contraction = 0
    highest_angular_momentum = 0
    number_of_contracted_shells = 0

    for i, atoms in enumerate(basis['atoms']):
        for shell in atoms['shells']:
            number_of_contracted_shells += 1
            st = shell['shell_type']
            shell_type.append(shell_type_list[st]['type'])
            n_primitives.append(len(shell['p_exponents']))
            atom_map.append(i+1)
            if highest_angular_momentum < shell_type_list[st]['angular_momentum']:
                highest_angular_momentum = shell_type_list[st]['angular_momentum']

            if len(shell['con_coefficients']) > largest_degree_of_contraction:
                    largest_degree_of_contraction = len(shell['con_coefficients'])

            for p in shell['p_exponents']:
                p_exponents.append(p)
            for c in shell['con_coefficients']:
                c_coefficients.append(c)
            for pc in shell['p_con_coefficients']:
                p_c_coefficients.append(pc)

    angstrom_to_bohr = 1/0.529177249
    coordinates_list = angstrom_to_bohr*structure.flatten()

    txt_fchk = '{}\n'.format('Extended Huckel calculation over filename')
    txt_fchk += 'SP        RHuckel                                                     {}\n'.format('STO-3G')
    txt_fchk += 'Number of atoms                            I               {}\n'.format(len(structure))
    txt_fchk += 'Charge                                     I               {}\n'.format(molecule.get_charge())
    txt_fchk += 'Multiplicity                               I               {}\n'.format(molecule.electronic_structure.
                                                                                         multiplicity)
    txt_fchk += 'Number of electrons                        I               {}\n'.format(number_of_electrons)
    txt_fchk += 'Number of alpha electrons                  I               {}\n'.format(alpha_electrons)
    txt_fchk += 'Number of beta electrons                   I               {}\n'.format(beta_electrons)
    txt_fchk += 'Number of basis functions                  I               {}\n'.format(number_of_basis_functions)

    txt_fchk += get_array_txt('Atomic numbers', 'I', atomic_numbers, row_size=6)
    txt_fchk += get_array_txt('Nuclear charges', 'R', atomic_numbers)
    txt_fchk += get_array_txt('Current cartesian coordinates', 'R', coordinates_list)

    txt_fchk += 'Number of contracted shells                I               {}\n'.format(number_of_contracted_shells)
    txt_fchk += 'Number of primitive shells                 I               {}\n'.format(np.sum(n_primitives))
    txt_fchk += 'Highest angular momentum                   I               {}\n'.format(highest_angular_momentum)
    txt_fchk += 'Largest degree of contraction              I               {}\n'.format(largest_degree_of_contraction)

    txt_fchk += get_array_txt('Shell types', 'I', shell_type, row_size=6)
    txt_fchk += get_array_txt('Number of primitives per shell', 'I', n_primitives, row_size=6)
    txt_fchk += get_array_txt('Shell to atom map', 'I', atom_map, row_size=6)
    txt_fchk += get_array_txt('Primitive exponents', 'R', p_exponents)
    txt_fchk += get_array_txt('Contraction coefficients', 'R', c_coefficients)
    txt_fchk += get_array_txt('P(S=P) Contraction coefficients', 'R', p_c_coefficients)
    # txt_fchk += get_array_txt('Coordinates of each shell', 'R', coor_shell) #
    # txt_fchk += get_array_txt('Overlap Matrix', 'R', overlap)
    #txt_fchk += get_array_txt('Core Hamiltonian Matrix', 'R', core_hamiltonian)
    txt_fchk += 'Total Energy                               R             {}\n'.format(total_energy)
    txt_fchk += get_array_txt('Alpha Orbital Energies', 'R', alpha_mo_energies)
    if beta_mo_energies is not None:
        txt_fchk += get_array_txt('Beta Orbital Energies', 'R', beta_mo_energies)
    # txt_fchk += get_array_txt('Total SCF Density', 'R', scf_density)
    txt_fchk += get_array_txt('Alpha MO coefficients', 'R', alpha_mo_coeff)
    if beta_mo_coeff is not None:
        txt_fchk += get_array_txt('Beta MO coefficients', 'R', beta_mo_coeff)

    return txt_fchk



from cosymlib import shape, tools
from cosymlib.symmetry import Symmetry
from cosymlib.symmetry.pointgroup import PointGroup
from cosymlib.tools import generate_connectivity_from_geometry
import numpy as np
from functools import wraps


def set_parameters(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        args[0]._symmetry.set_parameters(kwargs)
        return func(*args, **kwargs)
    return wrapper


class Geometry:
    """
    Main geometry class

    :param positions: Cartesian coordinates
    :type positions: list
    :param symbols: Atomic elements symbols
    :type symbols: list
    :param name: Geometry name
    :type name: str
    :param connectivity: Connectivity list
    :type connectivity: list
    :param connectivity_thresh: Connectivity threshold
    :type connectivity_thresh: float

    :example:

    .. code-block:: python

        water = Geometry(positions=[[0.0, 0.0, 0.0],
                                    [0.0, 1.0, 0.0],
                                    [0.0, 0.0, 1.0],
                         symbols=['O', 'H', 'H'],
                         name='water molecule',
                         connectivity=[[1, 2], [1, 3]])


    """
    def __init__(self,
                 positions,
                 symbols=(),
                 name='',
                 connectivity=None,
                 connectivity_thresh=1.2):

        # self._central_atom = None
        self._symbols = []
        self._positions = []
        self._atom_groups = list(symbols)
        self._name = name

# TODO: This is a mess!
        for symbol in symbols:
            try:
                int(symbol)
                self._symbols.append(tools.atomic_number_to_element(int(symbol)))
            except (ValueError, TypeError):
                self._symbols.append(symbol.capitalize())
                for ida, a in enumerate(symbol):
                    try:
                        int(a)
                        self._symbols[-1] = self._symbols[-1][:ida]
                        break
                    except (ValueError, TypeError, IndexError):
                        pass

        try:
            float(positions[0])
            for symbol in positions:
                self._positions.append(float(symbol))
            self._positions = list(chunks(self._positions, 3))
        except (ValueError, TypeError, IndexError):
            for symbol in positions:
                self._positions.append([float(j) for j in symbol])

        self._positions = np.array(self._positions)
        if connectivity is None:
            self._connectivity = generate_connectivity_from_geometry(self, thresh=connectivity_thresh)
        else:
            self._connectivity = connectivity

        self._shape = shape.Shape(self)
        self._symmetry = Symmetry(self)

    def __len__(self):
        return len(self._positions)

    @property
    def name(self):
        return self._name

    # TODO: 'symmetry' and 'shape' properties should be removed. Methods inside these
    # TODO: classes should be mirrored in geometry/molecule class. See get_symmetry_measure()

    @property
    def symmetry(self):
        return self._symmetry

    @property
    def shape(self):
        return self._shape

    def set_name(self, name):
        self._name = name

    def get_connectivity(self):
        """
        Get connectivity as a list of pairs of connected atoms

        :return: The connectivity
        :rtype: list
        """
        return self._connectivity

    def set_connectivity(self, connectivity):
        self._connectivity = connectivity
        self._symmetry._connectivity = connectivity

    def set_symbols(self, symbols):
        self._symbols = symbols
        self._symmetry._symbols = symbols

    def set_positions(self, central_atom=0):
        atom, self._positions = self._positions[central_atom], np.delete(self._positions, central_atom, 0)
        self._positions = np.insert(self._positions, len(self._positions), atom, axis=0)
        self._symmetry._positions= self._positions
        self._shape._positions= self._positions

    def get_positions(self):
        """
        Get the positions in Cartesian coordinates

        :return: the coordinates
        :rtype: list
        """
        return self._positions

    def get_n_atoms(self):
        """
        Get the number of atoms

        :return: number of atoms
        :rtype: int
        """
        return len(self.get_positions())

    def get_symbols(self):
        """
        Get the atomic elements symbols

        :return: the symbols
        :rtype: list
        """
        return self._symbols

    def generate_connectivity(self, thresh=1.2):
        self._connectivity = generate_connectivity_from_geometry(self, thresh=thresh)

    # Shape methods
    def get_shape_measure(self, shape_label, central_atom=0, fix_permutation=False):
        """
        Get the Shape measure

        :param shape_label: Reference shape label
        :type shape_label: str
        :param central_atom: the central atom position
        :type central_atom: int
        :param fix_permutation: Do not permute atoms
        :type fix_permutation: bool
        :return: The measure
        :rtype: float
        """
        return self._shape.measure(shape_label, central_atom=central_atom, fix_permutation=fix_permutation)

    def get_shape_structure(self, shape_label, central_atom=0, fix_permutation=False):
        from cosymlib.molecule.geometry import Geometry
        structure_coordinates = self._shape.structure(shape_label, central_atom=central_atom, fix_permutation=fix_permutation)

        return Geometry(symbols=self.get_symbols(),
                        positions=structure_coordinates,
                        name=self.name + '_structure')

    def get_path_deviation(self, shape_label1, shape_label2, central_atom=0):
        return self._shape.path_deviation(shape_label1, shape_label2, central_atom)

    def get_generalized_coordinate(self, shape_label1, shape_label2, central_atom=0):
        return self._shape.generalized_coordinate(shape_label1, shape_label2, central_atom)

    # Symmetry methods
    @set_parameters
    def get_symmetry_measure(self, label, central_atom=0, center=None, multi=1):
        """
        Get the symmetry measure

        :param label: Symmetry point group
        :type label: str
        :param central_atom: central atom position (0 if no central atom)
        :type central_atom: int
        :param center: center of the measure in Cartesian coordinates
        :type center: list
        :return: The symmetry measure
        :rtype: float
        """

        return self._symmetry.measure(label)

    @set_parameters
    def get_symmetry_nearest_structure(self, label, central_atom=0, center=None,  multi=1):
        """
        Returns the nearest ideal structure

        :param label: symmetry point group
        :type label: str
        :param central_atom: central atom position (0 if no central atom)
        :type central_atom: int
        :param center: center of the measure in Cartesian coordinates
        :type center: list
        :return: The structure
        :rtype: Geometry
        """
        return Geometry(symbols=self.get_symbols(),
                        positions=self._symmetry.nearest_structure(label),
                        name=self.name + '_nearest')
        # return self._symmetry.nearest_structure(label)

    @set_parameters
    def get_symmetry_optimum_axis(self, label, central_atom=0, center=None):
        return self._symmetry.optimum_axis(label)

    def get_pointgroup(self, tol=0.01):
        """
        Get the symmetry point group
        :param tol: The tolerance
        :type tol: float
        :return: The point group
        :rtype: str
        """
        return PointGroup(self, tolerance=tol).get_point_group()


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i+n]



import huckelpy
from huckelpy import file_io


class ExtendedHuckel:

    def __init__(self, geometry, charge=0):
        self._EH = huckelpy.ExtendedHuckel(geometry.get_positions(), geometry.get_symbols(), charge=charge)
        self._alpha_electrons = None
        self._beta_electrons = None
        self._total_electrons = self._EH.get_number_of_electrons()

    def get_mo_coefficients(self):
        return self._EH.get_eigenvectors()

    def get_basis(self):
        return self._EH.get_molecular_basis()

    def get_mo_energies(self):
        return self._EH.get_mo_energies()

    def get_multiplicity(self):
        return self._EH.get_multiplicity()

    def get_alpha_electrons(self):
        if self._alpha_electrons is None:
            self._alpha_electrons = self._total_electrons // 2 + self.get_multiplicity() - 1
        return self._alpha_electrons

    def get_beta_electrons(self):
        return self._total_electrons - self._alpha_electrons

    def build_fchk_file(self, name):
        txt_fchk = file_io.build_fchk(self._EH)
        open(name + '.fchk', 'w').write(txt_fchk)



import numpy as np
from warnings import warn
from cosymlib.symmetry.pointgroup.operations import rotation_matrix


def print_symmetry_labels():
    operations = {'E': 'Identity Symmetry',
                  'Ci': 'Inversion Symmetry Group',
                  'Cs': 'Reflection Symmetry Group',
                  'Cn': 'Rotational Symmetry Group (n: rotation order)',
                  'Sn': 'Rotation-Reflection Symmetry Group (n: rotation-reflection order)'}

    print('Available symmetry groups\n')
    for k, label in operations.items():
        print('{:6} {}'.format(k, label))

# def get_group_num_from_label(label):
#
#     operations = {'Cn':  '',
#                   'CnH': [2, ],
#                   'CnV': [3, ],
#                   'Ci':  [0, 2],
#                   'Cs':  [0, 3],
#                   'Cinf':[9, 1],
#                   'Dn':  [4, ],
#                   'DnH': [5, ],
#                   'DnD': [6, ],
#                   'Dinf':[9, 2],
#                   'Sn':  [7, ],
#                   'T':   [8, 1],
#                   'Th':  [8, 2],
#                   'Td':  [8, 3],
#                   'O':   [8, 4],
#                   'Oh':  [8, 5],
#                   'I':   [8, 6],
#                   'Ih':  [8, 7],
#                   }


def orthogonal_c4(c3, c4):
    def calculate_angle(axis1, axis2):
        axis1 /= np.linalg.norm(axis1)
        axis2 /= np.linalg.norm(axis2)
        return np.arccos(np.dot(axis1, axis2) / (np.linalg.norm(axis1) * np.linalg.norm(axis2))) * 180 / np.pi

    perfect_c3_c4 = 125.26438968275465
    c3 /= np.linalg.norm(c3)
    c4 /= np.linalg.norm(c4)
    angle_c3_c4 = calculate_angle(c3, c4)
    diff_angle = perfect_c3_c4 - angle_c3_c4
    if abs(diff_angle) <= 1E-8:
        new_c4 = c4
    else:
        new_c4 = np.dot(rotation_matrix(np.cross(c3, c4), diff_angle), c4)
        warn('ChangedAxisWarning: Set C4 axis to: {}'.format(new_c4))

    rot_mat = rotation_matrix(c3, 120.0)
    c4_2 = np.dot(rot_mat, new_c4)
    return new_c4, c4_2/np.linalg.norm(c4_2)


#!/usr/bin/python3
import sys
from os.path import expanduser
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5 import QtOpenGL
from PyQt5.Qt3DExtras import Qt3DWindow, QFirstPersonCameraController
from PyQt5.Qt3DCore import QEntity
from cosymlib.shape import tools
from cosymlib import file_io, shape
import numpy as np

home = expanduser("~")


class Execute(QPushButton):
    def __init__(self, title, parent):
        super().__init__(title, parent)

    def mousePressEvent(self, e):

        super().mousePressEvent(e)

        if e.button() == Qt.LeftButton:

                reply = QMessageBox.question(self, 'Message',
                                             "Run?", QMessageBox.Yes |
                                             QMessageBox.No, QMessageBox.No)

                if reply == QMessageBox.Yes:
                    quit()


class Exit(QPushButton):
    def __init__(self, title, parent):
        super().__init__(title, parent)

    def mousePressEvent(self, e):

        super().mousePressEvent(e)

        if e.button() == Qt.LeftButton:

                reply = QMessageBox.question(self, 'Message',
                                             "Are you sure to quit?", QMessageBox.Yes |
                                             QMessageBox.No, QMessageBox.No)

                if reply == QMessageBox.Yes:
                    quit()


class MainWindow(QWidget):

    def __init__(self):
        super().__init__()

        self.filename = None
        self.directory = None
        self.central_atom = 0
        self.initUI()

    def initUI(self):

        x = 500
        y = 300
        self.resize(x, y)
        self.center()

        # Input file
        file = QPushButton('Input file', self)
        file.move(x-100, -4)
        file.clicked.connect(self.open)
        self.file_text = QLineEdit(self)
        self.file_text.move(2, 2)
        self.file_text.resize(x - 150, 20)

        self.save_button = QPushButton('Save file', self)
        self.save_button.move(x-100, 20)
        self.save_button.setEnabled(False)
        self.save_button.clicked.connect(self.save)

        self.savefile_text = QLineEdit(self)
        self.savefile_text.move(2, 25)
        self.savefile_text.resize(x - 150, 20)

        # Label selection
        self.central_atom_name = QCheckBox('Central atom', self)
        self.central_atom_name.move(30, 60)
        self.central_atom_name.toggle()
        self.central_atom_name.stateChanged.connect(self.changeVertices)
        self.central_atom_box = QLineEdit(self)
        self.central_atom_box.move(150, 60)
        self.central_atom_box.resize(30, 30)
        self.central_atom_box.textChanged.connect(self.set_central_atom)

        labels_name = QLabel('Label :', self)
        labels_name.move(200, 60)
        self.labels_box = QComboBox(self)
        self.labels_box.move(250, 60)
        self.label = str(self.labels_box.currentText())
        self.labels_box.activated[str].connect(self.labels_on)

        self.results = QPlainTextEdit(self)
        self.results.setReadOnly(True)
        self.results.setStyleSheet("""QPlainTextEdit {background-color: #333; color: #00FF00; font-family: Courier;}""")
        self.results.move(30, 100)
        self.results.resize(250, 175)

        execute = QPushButton("Run")
        exit = Exit("Exit", self)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(execute)
        hbox.addWidget(exit)

        vbox = QVBoxLayout()
        vbox.addStretch(1)
        vbox.addLayout(hbox)

        self.setLayout(vbox)
        self.setWindowTitle('SHAPE')
        execute.clicked.connect(self.execute)

    def changeVertices(self, state):

        self.labels_box.clear()
        if state:
            self.central_atom_box.setReadOnly(False)
            if self.filename:
                self.labels_on(tools.get_structure_references(self.geometries[0].get_n_atoms() - 1))
        elif not state:
            # self.central_atom_box.clear()
            self.set_central_atom(0)
            self.central_atom_box.setReadOnly(True)
            if self.filename:
                self.labels_on(tools.get_structure_references(self.geometries[0].get_n_atoms()))

    def labels_on(self, vertices):
        for n in vertices:
            self.labels_box.addItem(n)
        self.label = str(self.labels_box.currentText())

    def open(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        self.filename, _ = QFileDialog.getOpenFileName(self, "Open", "", "All Files (*)", options=options)

        self.file_text.setText(str(self.filename))
        if self.filename:
            self.geometries = file_io.read_generic_structure_file(self.filename, read_multiple=True)
            self.changeVertices(self.central_atom_name.isChecked())
            # SecondWindow(self, local_geometry=[self.x(), self.y()]).show()

    def save(self):
        self.directory = QFileDialog.getExistingDirectory(self, "Select Directory")
        self.savefile_text.setText(str(self.directory))
        print(self.directory)
        # measures = [[shape.Shape(geometry).measure(self.label, self.central_atom)
        #              for geometry in self.geometries]]
        # names = [geometry.get_name() for geometry in self.geometries]
        # file_io.shape2file.write_shape_measure_data(measures, names, [self.label])

    def set_central_atom(self, text):
        try:
            self.central_atom = int(text)
        except ValueError:
            self.results.setPlainText('WARNING. Central atom must be an integer')
            self.results.repaint()
            # print('WARNING. Central atom must be an integer')

    def execute(self):
        if self.filename is None:
            self.results.setPlainText('WARNING. No input file selected')
            self.results.repaint()
            # print('WARNING. No input file selected, choose one before running')
        else:
            # if self.directory is None:
            text = "SHAPE's measure " + self.label + "\n"
            for geometry in self.geometries:
                text += '{:10} {:10.3f}'.format(geometry.name,
                                                shape.Shape(geometry).measure(self.label, self.central_atom)) + '\n'

            self.results.setPlainText(text)
            self.results.repaint()
            self.save_button.setEnabled(True)
            # else:
            #     measures = [[shape.Shape(geometry).measure(self.label, self.central_atom)
            #                 for geometry in self.geometries]]
            #     names = [geometry.get_name() for geometry in self.geometries]
            #     file_io.shape2file.write_shape_measure_data(measures, names, [self.label])

    def center(self):
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())


class SecondWindow(QMainWindow):
    def __init__(self, parent=None, local_geometry=[0, 0]):
        # super(SecondWindow, self).__init__(parent)
        QWidget.__init__(self, parent)
        self.setGeometry(local_geometry[0] + 500, local_geometry[1], 300, 400)
        self.setWindowTitle('Molecule View')
        self.pen = QPen(QColor(0,0,0))                      # set lineColor
        self.pen.setWidth(3)                                            # set lineWidth
        self.brush = QBrush(QColor(255,255,255,255))        # set fillColor
        self.polygon = self.createpoly(8, 150, 0)                         # polygon with n points, radius, angle of the first point


    def createpoly(self, n, r, s):
            polygon = QPolygonF()
            w = 360 / n  # angle per step
            for i in range(n):  # add the points of polygon
                t = w * i + s
                x = r * np.cos(np.radians(t))
                y = r * np.sin(np.radians(t))
                polygon.append(QPointF(self.width() / 2 + x, self.height() / 2 + y))

            return polygon

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setPen(self.pen)
        painter.setBrush(self.brush)
        painter.drawPolygon(self.polygon)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = MainWindow()
    ex.show()
    sys.exit(app.exec_())


HARTREE_TO_EV = 27.21139563180092


class ElectronicStructure:
    """
    Main Electronic structure class

    :param basis: The basis set
    :type basis: dict
    :param orbital_coefficients: Molecular orbital coefficients
    :type orbital_coefficients: list
    # :param charge: The charge
    # :type charge: int
    :param multiplicity: The multiplicity
    :type multiplicity: int
    :param alpha_energies: Alpha molecular orbital energies in Hartree
    :type alpha_energies: list
    :param beta_energies: Beta molecular orbital energies in Hartree
    :type beta_energies: list
    :param alpha_occupancy: Occupancy of alpha orbitals
    :type alpha_occupancy: list
    :param beta_occupancy: Occupancy of beta orbitals
    :type beta_occupancy: list

    """
    def __init__(self,
                 basis,
                 orbital_coefficients,
                 multiplicity=None,
                 alpha_energies=None,
                 beta_energies=None,
                 alpha_occupancy=None,
                 beta_occupancy=None):

        self._multiplicity = multiplicity
        self._basis = basis
        self._Ca = orbital_coefficients[0]
        if len(orbital_coefficients[1]) == 0:
            self._Cb = orbital_coefficients[0]
        else:
            self._Cb = orbital_coefficients[1]

        self._alpha_energies = [energy * HARTREE_TO_EV for energy in alpha_energies]
        if len(beta_energies) == 0:
            self._beta_energies = [energy * HARTREE_TO_EV for energy in alpha_energies]
        else:
            self._beta_energies = [energy * HARTREE_TO_EV for energy in beta_energies]

        self._alpha_occupancy = alpha_occupancy
        self._beta_occupancy = beta_occupancy
        self._occupancy_consistency()
        self._total_electrons = sum(self._alpha_occupancy) + sum(self._beta_occupancy)
        s = abs(sum([a_electron - self._beta_occupancy[ida] for ida, a_electron in enumerate(self._alpha_occupancy)]))
        self._s2 = s * (s + 1)
        if self._multiplicity is None:
            self._recalculate_spin()

    def set_occupancy(self, occupancy):
        self.set_alpha_occupancy(occupancy)
        self.set_beta_occupancy(occupancy)

    def set_alpha_occupancy(self, occupancy):
        if len(occupancy) > len(self._Ca):
            raise IndexError('Alpha occupancy list do not match with molecular orbitals')
        self._alpha_occupancy = occupancy
        self._occupancy_consistency()
        self._recalculate_spin()

    def set_beta_occupancy(self, occupancy):
        if len(occupancy) > len(self._Cb):
            raise IndexError('Beta occupancy list do not match with molecular orbitals')
        self._beta_occupancy = occupancy
        self._occupancy_consistency()
        self._recalculate_spin()

    def _recalculate_spin(self):
        s = abs(sum([a_electron - self._beta_occupancy[ida] for ida, a_electron in enumerate(self._alpha_occupancy)]))
        self._multiplicity = s + 1
        self._s2 = s*(s + 1)

    def _occupancy_consistency(self):
        if len(self._alpha_occupancy) < len(self._Ca):
            for i in range(len(self._Ca) - len(self._alpha_occupancy)):
                self._alpha_occupancy.append(0)
        if len(self._beta_occupancy) < len(self._Cb):
            for i in range(len(self._Cb) - len(self._beta_occupancy)):
                self._beta_occupancy.append(0)

    @property
    def multiplicity(self):
        return self._multiplicity

    @property
    def s2(self):
        return self._s2

    @property
    def basis(self):
        return self._basis

    @property
    def coefficients_a(self):
        return self._Ca

    @property
    def coefficients_b(self):
        return self._Cb

    @property
    def alpha_energies(self):
        return self._alpha_energies

    @property
    def beta_energies(self):
        return self._beta_energies

    @property
    def alpha_occupancy(self):
        return self._alpha_occupancy

    @property
    def beta_occupancy(self):
        return self._beta_occupancy

    @property
    def alpha_electrons(self):
        return sum(self._alpha_occupancy)

    @property
    def beta_electrons(self):
        return sum(self._beta_occupancy)



from cosymlib.shape import shp, tools
import numpy as np


def _get_key(central_atom, reference, fix_permutation=False, reference_2=''):
    if isinstance(reference, str):
        label_key = reference.lower()
    else:
        label_key = np.array2string(reference.get_positions(), precision=10)

    if isinstance(reference_2, str):
        label2_key = reference_2.lower()
    else:
        label2_key = np.array2string(reference_2.get_positions(), precision=10)

    central_atom_key = int(central_atom)
    fix_permutation_key = str(bool(fix_permutation))

    return central_atom_key, label_key, label2_key, fix_permutation_key


class Shape:
    """
    Shape main class

    :param structure: a geometry, molecule or array type object
    :type structure: Geometry, Molecule, np.array
    """
    def __init__(self, structure):

        # Allow geometry or molecule to be imported instead of crude Cartesian coordinates
        try:
            self._coordinates = structure.get_positions()
        except AttributeError:
            # Try to get from numpy array
            self._coordinates = structure

        self._coordinates = np.ascontiguousarray(self._coordinates)

        self._measures = {}
        self._structures = {}
        self._test_structures = {}
        self._path_deviation = {}
        self._gen_coord = {}

    def measure(self, reference, central_atom=0, fix_permutation=False):
        """
        Get shape measure

        :param reference: Reference shape label or Geometry object
        :type reference: str
        :param central_atom: Central atom position
        :type central_atom: int
        :param fix_permutation: Do not permute atoms during shape calculations
        :type fix_permutation: bool
        :return: The measure
        :rtype: float
        """
        key = _get_key(central_atom, reference, fix_permutation=fix_permutation)
        if key not in self._measures:
            if isinstance(reference, str):
                reference_structure = tools.get_reference_structure(reference, central_atom)
            else:
                reference_structure = reference
            reference_coordinates = reference_structure.get_positions()

            if len(self._coordinates) != len(reference_coordinates):
                raise Exception('Reference and input structures have different number of atoms')

            if fix_permutation:
                self._measures[key] = shp.cshm_fix(self._coordinates, reference_coordinates, central_atom)
            else:
                self._measures[key] = shp.cshm(self._coordinates, reference_coordinates, central_atom)

        return self._measures[key]

    def structure(self, reference, central_atom=0, fix_permutation=False):
        """
        Get the nearest structure to reference

        :param reference: Reference shape label or Geometry object
        :type reference: str, Geometry
        :param central_atom: Central atom position
        :type central_atom: int, Geometry
        :param fix_permutation: Do not permute atoms during shape calculations
        :type fix_permutation: bool
        :return: The structure
        :rtype: Structure
        """
        key = _get_key(central_atom, reference, fix_permutation=fix_permutation)
        if key not in self._structures:
            if isinstance(reference, str):
                reference_structure = tools.get_reference_structure(reference, central_atom)
            else:
                reference_structure = reference

            reference_coordinates = reference_structure.get_positions()

            if len(self._coordinates) != len(reference_coordinates):
                raise Exception('Reference and input structures have different number of atoms')

            if fix_permutation:
                self._structures[key], self._measures[key] = shp.poly_fix(self._coordinates,
                                                                          reference_coordinates,
                                                                          central_atom)
            else:
                self._structures[key], self._measures[key] = shp.poly(self._coordinates,
                                                                      reference_coordinates,
                                                                      central_atom)

        return self._structures[key]

    def path_deviation(self, shape_label1, shape_label2, central_atom=0):
        """
        Get the path deviation

        :param shape_label1: First shape reference label or Geometry object
        :type shape_label1: str, Geometry
        :param shape_label2: Second shape reference label or Geometry object
        :type shape_label2: str, Geometry
        :param central_atom: Central atom position
        :type central_atom: int
        :return: The path deviation
        :rtype: float
        """
        # TODO: Someone improve the description of this function

        key = _get_key(central_atom, shape_label1, reference_2=shape_label2)
        if key not in self._path_deviation:
            Sx = self.measure(shape_label1, central_atom)
            Sy = self.measure(shape_label2, central_atom)
            new_theta = np.arcsin(np.sqrt(Sx) / 10) + np.arcsin(np.sqrt(Sy) / 10)
            if isinstance(shape_label1, str):
                structure_a = tools.get_reference_structure(shape_label1, central_atom=central_atom)
            else:
                structure_a = shape_label1

            theta = np.arcsin(np.sqrt(Shape(structure_a).measure(shape_label2, central_atom=structure_a.get_n_atoms())) / 10)
            self._path_deviation[key] = ((new_theta / theta) - 1) * 100

        return self._path_deviation[key]

    def generalized_coordinate(self, shape_label1, shape_label2, central_atom=0):
        """
        Get the generalized coordinate

        :param shape_label1: First shape reference label or Geometry object
        :type shape_label1: str, Geometry
        :param shape_label2: Second shape reference label or Geometry object
        :type shape_label2: str, Geometry
        :param central_atom: Central atom position
        :type central_atom: int
        :return: The generalized coordinate
        """

        key = _get_key(central_atom, shape_label1, reference_2=shape_label2)
        if key not in self._gen_coord:
            Sq = self.measure(shape_label1, central_atom)
            if isinstance(shape_label1, str):
                structure_a = tools.get_reference_structure(shape_label1, central_atom=central_atom)
            else:
                structure_a = shape_label1

            theta = np.arcsin(np.sqrt(Shape(structure_a).measure(shape_label2, central_atom=structure_a.get_n_atoms())) / 10)
            self._gen_coord[key] = round(100 * np.arcsin(np.sqrt(Sq) / 10) / theta, 1)

        return self._gen_coord[key]



from cosymlib.shape import tools, Shape
import numpy as np
from cosymlib.molecule.geometry import Geometry


def get_shape_map(shape_label1, shape_label2, num_points=20):

    path_structures = []
    if isinstance(shape_label1, Geometry):
        coordinates_a = shape_label1.get_positions()
    else:
        coordinates_a = tools.get_reference_structure(shape_label1, central_atom=1).get_positions()
    coordinates_b = Shape(coordinates_a).structure(shape_label2, central_atom=len(coordinates_a))
    coordinates_b = np.concatenate((coordinates_b[1:], [coordinates_b[0]]))


    measure_label1 = []
    measure_label2 = []
    for i in range(0, num_points+1):
        structure_along_path = []
        for ida, atom in enumerate(coordinates_a):
            structure_along_path.append(atom + (coordinates_b[ida] - atom)*i/num_points)

        measure_label1.append(Shape(structure_along_path).measure(shape_label1, central_atom=len(coordinates_a)))
        measure_label2.append(Shape(structure_along_path).measure(shape_label2, central_atom=len(coordinates_a)))
        path_structures.append(structure_along_path)

    return measure_label1, measure_label2, path_structures



from cosymlib.shape import maps
import numpy as np
import sys


def plot_minimum_distortion_path_shape(shape_label1, shape_label2, num_points=20, output=sys.stdout, show_plot=True):
    import matplotlib.pyplot as plt

    path = get_shape_path(shape_label1, shape_label2, num_points)

    shape_map_txt = " {:6} {:6}\n".format(shape_label1, shape_label2)
    for idx, value in enumerate(path[0]):
        shape_map_txt += '{:6.3f}, {:6.3f}'.format(path[0][idx], path[1][idx])
        shape_map_txt += '\n'

    print(shape_map_txt)
    if show_plot:
        plt.plot(path[0], path[1], 'k', linewidth=2.0)
        plt.xlabel(shape_label1)
        plt.ylabel(shape_label2)
        plt.show()


def get_shape_path(shape_label1, shape_label2, num_points):
    return maps.get_shape_map(shape_label1, shape_label2, num_points)


def plot_molecular_orbital_diagram(molecule, wfnsym, mo_range=None):
    import matplotlib.pyplot as plt

    labels = wfnsym.IRLab
    if mo_range is not None:
        ird_a_max = [np.argmax(ird_a_orb) for ird_a_orb in wfnsym.mo_IRd_a][mo_range[0]:mo_range[1]]
        energies = molecule.electronic_structure.alpha_energies[mo_range[0]:mo_range[1]]
    else:
        ird_a_max = [np.argmax(ird_a_orb) for ird_a_orb in wfnsym.mo_IRd_a]
        energies = molecule.electronic_structure.alpha_energies

    ax1 = plt.axes()
    ax1.axes.get_xaxis().set_visible(False)  # Hide x axis
    # ax1.axes.get_yaxis().set_visible(True)

    degeneracy = [[energies[0]]]
    for energy in energies[1:]:
        if abs(energy - degeneracy[-1][-1]) < 1e-3:
            degeneracy[-1].append(energy)
        else:
            degeneracy.append([energy])

    max_value = 5e-3
    x_center = []
    for ix in degeneracy:
        if len(ix) == 1:
            x_center.append([0])
        else:
            x_center.append(np.linspace(-max_value, max_value, len(ix)))
    x_center = [y for x in x_center for y in x]

    plt.scatter(x_center, energies, s=500, marker="_", linewidth=3)
    for i in range(len(energies)):
        plt.text(-max_value * 2, energies[i], labels[ird_a_max[i]])

    plt.show()


def swap_vectors(v1, v2, position):
    vector1 = v1.get_copy()
    vector2 = v2.get_copy()
    for i in range(len(v1)):
        if i >= position:
            vector1[i] = v2[i]
            vector2[i] = v1[i]
    return vector1, vector2


def plot_symmetry_energy_evolution(molecules, wfnsym, mo_range=None):
    import matplotlib.pyplot as plt

    energies = []
    ird_a_max = []
    for idm, molecule in enumerate(molecules):
        labels = wfnsym[idm].IRLab
        if mo_range is not None:
            ird_a_max.append(np.array([np.argmax(ird_a_orb) for ird_a_orb in wfnsym[idm].mo_IRd_a]
                                      [mo_range[0]:mo_range[1]]))
            energies.append(molecule.electronic_structure.alpha_energies[mo_range[0]:mo_range[1]])
        else:
            ird_a_max.append(np.array([np.argmax(ird_a_orb) for ird_a_orb in wfnsym[idm].mo_IRd_a]))
            energies.append(molecule.electronic_structure.alpha_energies)

    energies_x_orbital = np.array(energies).T
    ird_a_x_orbital = np.array(ird_a_max).T

    for i in range(len(ird_a_x_orbital)):
        for j in range(len(ird_a_x_orbital[i])):
            if j == 0:
                old_ird = ird_a_x_orbital[i][0]
            else:
                if old_ird != ird_a_x_orbital[i][j]:
                    for k in range(len(ird_a_x_orbital) - i):
                        if old_ird == ird_a_x_orbital[k + i][j]:
                            ird_a_x_orbital[i], ird_a_x_orbital[k + i] = swap_vectors(ird_a_x_orbital[i],
                                                                                      ird_a_x_orbital[k + i], j)
                            energies_x_orbital[i], energies_x_orbital[k + i] = swap_vectors(energies_x_orbital[i],
                                                                                            energies_x_orbital[k + i],
                                                                                            j)
                            break
            old_ird = ird_a_x_orbital[i][j]

    for ide, energy in enumerate(energies_x_orbital):
        x = np.arange(len(energy))
        plt.plot(x, energy, marker='_')
        for i in range(len(energy)):
            plt.text(x[i], energy[i] + abs(energy[i])*0.001, labels[ird_a_x_orbital[ide][i]])

    plt.show()



from wfnsympy import WfnSympy
from symgroupy import Symgroupy
import numpy as np


def _get_key_symgroup(label, center, central_atom, connectivity, multi, connect_thresh):
    group_key = label.lower()
    center_key = ' '.join(['{:10.8f}'.format(n) for n in center]) if center is not None else None
    connectivity_key = np.array2string(np.array(connectivity), precision=10) if connectivity is not None else None
    multi_key = int(multi)
    central_atom_key = int(central_atom)
    connect_thresh_key = '{:10.8f}'.format(connect_thresh)
    return group_key, center_key, central_atom_key, connectivity_key, multi_key, connect_thresh_key


def _get_key_wfnsym(group, vector_axis1, vector_axis2, center, alpha_occupancy, beta_occupancy):
    group_key = group.lower()
    vec1_key = ' '.join(['{:10.8f}'.format(n) for n in vector_axis1]) if vector_axis1 is not None else None
    vec2_key = ' '.join(['{:10.8f}'.format(n) for n in vector_axis2]) if vector_axis2 is not None else None
    center_key = ' '.join(['{:10.8f}'.format(n) for n in center]) if center is not None else None
    alpha_occupancy_key = ' '.join(['{:10.8f}'.format(n) for n in alpha_occupancy]) if alpha_occupancy is not None else None
    beta_occupancy_key = ' '.join(['{:10.8f}'.format(n) for n in beta_occupancy]) if beta_occupancy is not None else None
    return group_key, vec1_key, vec2_key, center_key, alpha_occupancy_key, beta_occupancy_key


class Symmetry:
    """
    Symmetry main class

    :param structure: a geometry, molecule or array type object
    :type structure: Geometry, Molecule, np.array
    :param central_atom: central atom position
    :type central_atom: int
    :param center: center of symmetry in Cartesian coordinates. By default center is optimized
    :type center: list
    :param connect_thresh: Connectivity threshold (Ionic radius is used as reference)
    :type connect_thresh: float
    :param multi: Number of symmetry axis to find
    :type multi: int
    :param axis: Main symmetry axis (If None, then optimize)
    :type axis: list
    :param axis2: secondary symmetry axis (If None, then optimize)
    :type axis2: list
    """

    def __init__(self,
                 structure,
                 central_atom=0,
                 center=None,
                 connect_thresh=1.2,
                 multi=1,
                 axis=None,
                 axis2=None,
                 ):

        try:
            # Interpret as Geometry or molecule
            self._coordinates = structure.get_positions()
            self._symbols = structure.get_symbols()
            self._connectivity = structure.get_connectivity()
        except AttributeError:
            # Interpret as numpy array of coordinates or list of lists
            self._coordinates = structure
            self._symbols = None
            self._connectivity = None

        # If molecule object add electronic structure
        try:
            self._electronic_structure = structure.electronic_structure
        except AttributeError:
            self._electronic_structure = None

        self._central_atom = central_atom
        self._center = center
        self._connect_thresh = connect_thresh
        self._multi = multi
        self._axis = axis
        self._axis2 = axis2
        self._results = {}

    # Modifier methods
    def set_parameters(self, parameters_dict):
        """
        Set symmetry calculation related parameters

        :param parameters_dict: parameters in dictionary
        :type parameters_dict: dict
        """
        for name, value in parameters_dict.items():
            setattr(self, '_' + name, value)

    def set_electronic_structure(self, electronic_structure):
        self._electronic_structure = electronic_structure

    def _get_symgroup_results(self, group):

        """
        # Temporal interface
        if central_atom is not None:
            self._central_atom = central_atom

        self._multi = multi
        self._center = center
        self._connect_thresh = connect_thresh
        """

        # Crude calculation call methods
        key = _get_key_symgroup(group, self._center, self._central_atom, self._connectivity, self._multi, self._connect_thresh)
        if key not in self._results:
            self._results[key] = Symgroupy(self._coordinates,
                                           group=group,
                                           labels=self._symbols,
                                           central_atom=self._central_atom,
                                           multi=self._multi,
                                           center=self._center,
                                           connectivity=self._connectivity,
                                           connect_thresh=self._connect_thresh,
                                           permutation=None)

        return self._results[key]

    def _get_wfnsym_results(self, group):
        if self._electronic_structure is None:
            raise Exception('Electronic structure not set')

        key = _get_key_wfnsym(group, self._axis, self._axis2, self._center, self._electronic_structure.alpha_occupancy,
                              self._electronic_structure.beta_occupancy)

        if key not in self._results:
            self._results[key] = WfnSympy(coordinates=self._coordinates,
                                          symbols=self._symbols,
                                          basis=self._electronic_structure.basis,
                                          center=self._center,
                                          axis=self._axis,
                                          axis2=self._axis2,
                                          alpha_mo_coeff=self._electronic_structure.coefficients_a,
                                          beta_mo_coeff=self._electronic_structure.coefficients_b,
                                          group=group.upper(),
                                          alpha_occupancy=self._electronic_structure.alpha_occupancy,
                                          beta_occupancy=self._electronic_structure.beta_occupancy)
        return self._results[key]

    ##########################################
    #       Structure symmetry methods       #
    ##########################################

    def measure(self, label):
        """
        Get symmetry measure

        :param label: Point group label
        :type label: str
        :return: The measure
        :rtype: float
        """
        return self._get_symgroup_results(label).csm

    def nearest_structure(self, label):
        """
        Get nearest structure

        :param label: Point group label
        :type label: str
        :return: The structure
        :rtype: Structure
        """
        # TODO: Improve this docstring

        return self._get_symgroup_results(label).nearest_structure

    def optimum_axis(self, label):
        """
        Get the optimum main symmetry axis

        :param label: Point group label
        :type label: str
        :return: The axis
        :rtype: list
        """
        return self._get_symgroup_results(label).optimum_axis

    def optimum_permutation(self, label):
        """
        Get the optimum atoms permutation

        :param label: point group label
        :return: The permutation
        :rtype: list
        """
        return self._get_symgroup_results(label).optimum_permutation

    def reference_axis(self, label):
        """
        Get reference axis

        :param label: point group label
        :type label: str
        :return: The axis
        :rtype: list
        """
        return self._get_symgroup_results(label).reference_axis

    def cms_multi(self, label, multi=1):
        """
        Get symmetry measure of the optimum N axis

        :param label: point group label
        :type label: str
        :param multi: number of axis
        :type multi: int
        :return: The measures
        :rtype: list
        """
        self._multi = multi
        return self._get_symgroup_results(label).cms_multi

    def axis_multi(self, label, multi=1):
        """
        Get the optimum N axis

        :param label: point group label
        :type label: str
        :param multi: number of axis
        :type multi: int
        :return: List of axis
        :rtype: list
        """
        self._multi = multi
        return self._get_symgroup_results(label).axis_multi

    ##########################################
    #       Electronic symmetry methods      #
    ##########################################

    # TODO: Consider to migrate this methods data to pandas tabular structures
    def mo_irreducible_representations(self, group):
        results = self._get_wfnsym_results(group)

        return {'labels': results.IRLab,
                'alpha': results.mo_IRd_a,
                'beta': results.mo_IRd_b}

    def wf_irreducible_representations(self, group):
        results = self._get_wfnsym_results(group)

        return {'labels': results.SymLab,
                'alpha': results.wf_IRd_a,
                'beta': results.wf_IRd_b,
                'total': results.wf_IRd}

    def mo_overlaps(self, group):
        results = self._get_wfnsym_results(group)

        return {'labels': results.SymLab,
                'alpha': results.mo_SOEVs_a,
                'beta': results.mo_SOEVs_b}

    def wf_overlaps(self, group):
        results = self._get_wfnsym_results(group)

        return {'labels': results.SymLab,
                'alpha': results.wf_SOEVs_a,
                'beta': results.wf_SOEVs_b,
                'total': results.wf_SOEVs}

    def symmetry_matrix(self, group):
        results = self._get_wfnsym_results(group)
        return {'labels': results.SymLab,
                'matrix': results.SymMat}

    def wf_measure(self, group):
        results = self._get_wfnsym_results(group)
        return {'labels': results.SymLab,
                'csm': results.csm_coef,
                'grim': results.grim_coef}

    def wf_ideal_group_table(self, group):
        results = self._get_wfnsym_results(group)
        return {'ir_labels' : results.IRLab,
                'labels': results.SymLab,
                'table': results.ideal_gt}

    def dens_measure(self, group):
        results = self._get_wfnsym_results(group)
        return {'labels': results.SymLab,
                'csm': results.csm_dens,
                'csm_coef': results.csm_dens_coef}

    def axes(self, group):
        results = self._get_wfnsym_results(group)
        return {'center' : results.center,
                'axis': results.axis,
                'axis2': results.axis2}
